"""
Functions to do automatic visualization of Niimg-like objects
See http://nilearn.github.io/manipulating_images/input_output.html

Only matplotlib is required.
"""

# Author: Gael Varoquaux, Chris Filo Gorgolewski
# License: BSD

# Standard library imports
import os
import collections.abc
import functools
import numbers
import warnings
from distutils.version import LooseVersion

# Standard scientific libraries imports (more specific imports are
# delayed, so that the part module can be used without them).
import numpy as np
from scipy import ndimage
from scipy import sparse
from scipy import stats
from nibabel.spatialimages import SpatialImage

from ..signal import clean
from .._utils.numpy_conversions import as_ndarray
from .._utils.niimg import _safe_get_data

import matplotlib
import matplotlib.pyplot as plt
from matplotlib import gridspec as mgs

from .. import _utils
from .._utils.extmath import fast_abs_percentile
from .._utils.param_validation import check_threshold
from .._utils.ndimage import get_border_data
from ..datasets import load_mni152_template
from ..image import new_img_like, iter_img, get_data, math_img, resample_to_img
from ..input_data import NiftiMasker
from nilearn.image.resampling import reorder_img
from ..masking import compute_epi_mask, apply_mask
from .displays import get_slicer, get_projector
from . import cm


def show():
    """Show all the figures generated by nilearn and/or matplotlib.

    This function is equivalent to :func:`matplotlib.pyplot.show`,
    but is skipped on the 'Agg' backend where it has no effect other
    than to emit a warning.

    """
    if matplotlib.get_backend().lower() != 'agg':  # avoid warnings
        plt.show()


###############################################################################
# Core, usage-agnostic functions


def _get_colorbar_and_data_ranges(stat_map_data, vmax, symmetric_cbar, kwargs,
                                  force_min_stat_map_value=None):
    """Internal function for setting colormap and colorbar limits.

    Used by for plot_stat_map and plot_glass_brain.

    The limits for the colormap will always be set to range from -vmax to vmax.
    The limits for the colorbar depend on the symmetric_cbar argument, please
    refer to docstring of plot_stat_map.

    """
    if 'vmin' in kwargs:
        raise ValueError('this function does not accept a "vmin" '
                         'argument, as it uses a symmetrical range '
                         'defined via the vmax argument. To threshold '
                         'the plotted map, use the "threshold" argument')

    # make sure that the color range is symmetrical
    if vmax is None or symmetric_cbar in ['auto', False]:
        # Avoid dealing with masked_array:
        if hasattr(stat_map_data, '_mask'):
            stat_map_data = np.asarray(
                stat_map_data[np.logical_not(stat_map_data._mask)])
        stat_map_max = np.nanmax(stat_map_data)
        if force_min_stat_map_value is None:
            stat_map_min = np.nanmin(stat_map_data)
        else:
            stat_map_min = force_min_stat_map_value

    if symmetric_cbar == 'auto':
        symmetric_cbar = stat_map_min < 0 and stat_map_max > 0

    if vmax is None:
        vmax = max(-stat_map_min, stat_map_max)
    vmin = -vmax

    if not symmetric_cbar:
        negative_range = stat_map_max <= 0
        positive_range = stat_map_min >= 0
        if positive_range:
            cbar_vmin = 0
            cbar_vmax = None
        elif negative_range:
            cbar_vmax = 0
            cbar_vmin = None
        else:
            cbar_vmin = stat_map_min
            cbar_vmax = stat_map_max
    else:
        cbar_vmin, cbar_vmax = None, None
    return cbar_vmin, cbar_vmax, vmin, vmax


def _plot_img_with_bg(img, bg_img=None, cut_coords=None,
                      output_file=None, display_mode='ortho',
                      colorbar=False, figure=None, axes=None, title=None,
                      threshold=None, annotate=True,
                      draw_cross=True, black_bg=False,
                      vmin=None, vmax=None,
                      bg_vmin=None, bg_vmax=None, interpolation="nearest",
                      display_factory=get_slicer,
                      cbar_vmin=None, cbar_vmax=None,
                      brain_color=(0.5, 0.5, 0.5),
                      decimals=False,
                      **kwargs):
    """Internal function, please refer to the docstring of plot_img for
    parameters not listed below.

    Parameters
    ----------
    img : Niimg like object
        Image to plot.

    bg_vmin : float, optional
        vmin for bg_img.

    bg_vmax : float, optional
        vmax for bg_img.

    interpolation : string, optional
        Passed to the add_overlay calls.
        Default='nearest'.

    display_factory : function, optional
        Takes a display_mode argument and return a display class.
        Default=get_slicer.

    """
    show_nan_msg = False
    if vmax is not None and np.isnan(vmax):
        vmax = None
        show_nan_msg = True
    if vmin is not None and np.isnan(vmin):
        vmin = None
        show_nan_msg = True
    if show_nan_msg:
        nan_msg = ('NaN is not permitted for the vmax and vmin arguments.\n'
                   'Tip: Use np.nanmax() instead of np.max().')
        warnings.warn(nan_msg)

    if (isinstance(cut_coords, numbers.Number) and
            (display_mode == 'ortho' or display_mode == 'tiled')):
        raise ValueError(
            "The input given for display_mode='{0}' needs to be "
            "a list of 3d world coordinates in (x, y, z). "
            "You provided single cut, "
            "cut_coords={1}".format(display_mode, cut_coords))

    if img is not False and img is not None:
        img = _utils.check_niimg_3d(img, dtype='auto')
        data = _safe_get_data(img, ensure_finite=True)
        affine = img.affine

        if np.isnan(np.sum(data)):
            data = np.nan_to_num(data)

        # Deal with automatic settings of plot parameters
        if threshold == 'auto':
            # Threshold epsilon below a percentile value, to be sure that some
            # voxels pass the threshold
            threshold = fast_abs_percentile(data) - 1e-5

        img = new_img_like(img, as_ndarray(data), affine)

    display = display_factory(display_mode)(
        img,
        threshold=threshold,
        cut_coords=cut_coords,
        figure=figure, axes=axes,
        black_bg=black_bg,
        colorbar=colorbar,
        brain_color=brain_color,
        )

    if bg_img is not None:
        bg_img = _utils.check_niimg_3d(bg_img)
        display.add_overlay(bg_img,
                            vmin=bg_vmin, vmax=bg_vmax,
                            cmap=plt.cm.gray, interpolation=interpolation)

    if img is not None and img is not False:
        display.add_overlay(new_img_like(img, data, affine),
                            threshold=threshold, interpolation=interpolation,
                            colorbar=colorbar, vmin=vmin, vmax=vmax,
                            **kwargs)

    if annotate:
        display.annotate(decimals=decimals)
    if draw_cross:
        display.draw_cross()
    if title is not None and not title == '':
        display.title(title)
    if hasattr(display, '_cbar'):
        cbar = display._cbar
        _crop_colorbar(cbar, cbar_vmin, cbar_vmax)
    if output_file is not None:
        display.savefig(output_file)
        display.close()
        display = None
    return display


def _crop_colorbar(cbar, cbar_vmin, cbar_vmax):
    """Crop a colorbar to show from cbar_vmin to cbar_vmax.
    Used when symmetric_cbar=False is used.

    """
    if (cbar_vmin is None) and (cbar_vmax is None):
        return
    cbar_tick_locs = cbar.locator.locs
    if cbar_vmax is None:
        cbar_vmax = cbar_tick_locs.max()
    if cbar_vmin is None:
        cbar_vmin = cbar_tick_locs.min()
    new_tick_locs = np.linspace(cbar_vmin, cbar_vmax,
                                len(cbar_tick_locs))

    # matplotlib >= 3.2.0 no longer normalizes axes between 0 and 1
    # See https://matplotlib.org/3.2.1/api/prev_api_changes/api_changes_3.2.0.html
    # _outline was removed in
    # https://github.com/matplotlib/matplotlib/commit/03a542e875eba091a027046d5ec652daa8be6863
    # so we use the code from there
    if LooseVersion(matplotlib.__version__) >= LooseVersion("3.2.0"):
        cbar.ax.set_ylim(cbar_vmin, cbar_vmax)
        X, _ = cbar._mesh()
        X = np.array([X[0], X[-1]])
        Y = np.array([[cbar_vmin, cbar_vmin], [cbar_vmax, cbar_vmax]])
        N = X.shape[0]
        ii = [0, 1, N - 2, N - 1, 2 * N - 1, 2 * N - 2, N + 1, N, 0]
        x = X.T.reshape(-1)[ii]
        y = Y.T.reshape(-1)[ii]
        xy = (np.column_stack([y, x])
              if cbar.orientation == 'horizontal' else
              np.column_stack([x, y]))
        cbar.outline.set_xy(xy)
    else:
        cbar.ax.set_ylim(cbar.norm(cbar_vmin), cbar.norm(cbar_vmax))
        outline = cbar.outline.get_xy()
        outline[:2, 1] += cbar.norm(cbar_vmin)
        outline[2:6, 1] -= (1. - cbar.norm(cbar_vmax))
        outline[6:, 1] += cbar.norm(cbar_vmin)
        cbar.outline.set_xy(outline)

    cbar.set_ticks(new_tick_locs, update_ticks=True)


def plot_img(img, cut_coords=None, output_file=None, display_mode='ortho',
             figure=None, axes=None, title=None, threshold=None,
             annotate=True, draw_cross=True, black_bg=False, colorbar=False,
             resampling_interpolation='continuous',
             bg_img=None, vmin=None, vmax=None, **kwargs):
    """ Plot cuts of a given image (by default Frontal, Axial, and Lateral)

    Parameters
    ----------
    img : Niimg-like object
        See http://nilearn.github.io/manipulating_images/input_output.html

    cut_coords : None, a tuple of floats, or an integer, optional
        The MNI coordinates of the point where the cut is performed
        If display_mode is 'ortho' or 'tiled',
        this should be a 3-tuple: (x, y, z)
        For display_mode == 'x', 'y', or 'z', then these are the
        coordinates of each cut in the corresponding direction.
        If None is given, the cuts is calculated automaticaly.
        If display_mode is 'x', 'y' or 'z', cut_coords can be an integer,
        in which case it specifies the number of cuts to perform

    output_file : string, or None, optional
        The name of an image file to export the plot to. Valid extensions
        are .png, .pdf, .svg. If output_file is not None, the plot
        is saved to a file, and the display is closed.

    display_mode : {'ortho', 'tiled', 'x', 'y', 'z', 'yx', 'xz', 'yz'}
        Choose the direction of the cuts: 'x' - sagittal, 'y' - coronal,
        'z' - axial, 'ortho' - three cuts are performed in orthogonal
        directions, 'tiled' - three cuts are performed
        and arranged in a 2x2 grid.
        Default='ortho'.

    figure : integer or matplotlib figure, optional
        Matplotlib figure used or its number. If None is given, a
        new figure is created.

    axes : matplotlib axes or 4 tuple of float: (xmin, ymin, width, height), optional
        The axes, or the coordinates, in matplotlib figure space,
        of the axes used to display the plot. If None, the complete
        figure is used.

    title : string, optional
        The title displayed on the figure.

    threshold : a number, None, or 'auto', optional
        If None is given, the image is not thresholded.
        If a number is given, it is used to threshold the image:
        values below the threshold (in absolute value) are plotted
        as transparent. If auto is given, the threshold is determined
        magically by analysis of the image.

    annotate : boolean, optional
        If annotate is True, positions and left/right annotation
        are added to the plot. Default=True.

    decimals : integer, optional
        Number of decimal places on slice position annotation. If False (default),
        the slice position is integer without decimal point.

    draw_cross : boolean, optional
        If draw_cross is True, a cross is drawn on the plot to
        indicate the cut plosition. Default=True.

    black_bg : boolean, optional
        If True, the background of the image is set to be black. If
        you wish to save figures with a black background, you
        will need to pass "facecolor='k', edgecolor='k'"
        to matplotlib.pyplot.savefig. Default=False.

    colorbar : boolean, optional
        If True, display a colorbar on the right of the plots.
        Default=False.

    resampling_interpolation : str, optional
        Interpolation to use when resampling the image to the destination
        space. Can be "continuous" to use 3rd-order spline interpolation,
        or "nearest" to use nearest-neighbor mapping. "nearest" is faster
        but can be noisier in some cases. Default='continuous'.

    bg_img : Niimg-like object, optional
        See http://nilearn.github.io/manipulating_images/input_output.html
        The background image that the ROI/mask will be plotted on top of.
        If nothing is specified, no background image is plotted.

    vmin : float, optional
        Lower bound of the colormap. If `None`, the min of the image is used.

    vmax : float, optional
        Upper bound of the colormap. If `None`, the max of the image is used.

    kwargs : extra keyword arguments, optional
        Extra keyword arguments passed to matplotlib.pyplot.imshow.

    """  # noqa: E501
    display = _plot_img_with_bg(
        img, cut_coords=cut_coords,
        output_file=output_file, display_mode=display_mode,
        figure=figure, axes=axes, title=title,
        threshold=threshold, annotate=annotate,
        draw_cross=draw_cross,
        resampling_interpolation=resampling_interpolation,
        black_bg=black_bg, colorbar=colorbar,
        bg_img=bg_img, vmin=vmin, vmax=vmax, **kwargs)

    return display


###############################################################################
# Anatomy image for background

# A constant class to serve as a sentinel for the default MNI template
class _MNI152Template(SpatialImage):
    """This class is a constant pointing to the MNI152 Template
    provided by nilearn.

    """
    data = None
    affine = None
    vmax = None
    _shape = None
    # Having a header is required by the load_niimg function
    header = None

    def __init__(self, data=None, affine=None, header=None):
        # Comply with spatial image requirements while allowing empty init
        pass

    def load(self):
        if self.data is None:
            anat_img = load_mni152_template()
            anat_img = reorder_img(anat_img)
            data = get_data(anat_img)
            data = data.astype(np.float64)
            anat_mask = ndimage.morphology.binary_fill_holes(data > 0)
            data = np.ma.masked_array(data, np.logical_not(anat_mask))
            self._affine = anat_img.affine
            self.data = data
            self.vmax = data.max()
            self._shape = anat_img.shape

    @property
    def _data_cache(self):
        self.load()
        return self.data

    @property
    def _dataobj(self):
        self.load()
        return self.data

    def get_data(self):
        self.load()
        return self.data

    @property
    def affine(self):
        self.load()
        return self._affine

    def get_affine(self):
        self.load()
        return self._affine

    @property
    def shape(self):
        self.load()
        return self._shape

    def get_shape(self):
        self.load()
        return self._shape

    def __str__(self):
        return "<MNI152Template>"

    def __repr__(self):
        return "<MNI152Template>"


# The constant that we use as a default in functions
MNI152TEMPLATE = _MNI152Template()


def _load_anat(anat_img=MNI152TEMPLATE, dim='auto', black_bg='auto'):
    """ Internal function used to load anatomy, for optional diming
    """
    vmin = None
    vmax = None
    if anat_img is False or anat_img is None:
        if black_bg == 'auto':
            # No anatomy given: no need to turn black_bg on
            black_bg = False
        return anat_img, black_bg, vmin, vmax

    if anat_img is MNI152TEMPLATE:
        anat_img.load()
        # We special-case the 'canonical anat', as we don't need
        # to do a few transforms to it.
        vmin = 0
        vmax = anat_img.vmax
        if black_bg == 'auto':
            black_bg = False
    else:
        anat_img = _utils.check_niimg_3d(anat_img)
        # Clean anat_img for non-finite values to avoid computing unnecessary
        # border data values.
        data = _safe_get_data(anat_img, ensure_finite=True)
        anat_img = new_img_like(anat_img, data, affine=anat_img.affine)
        if dim or black_bg == 'auto':
            # We need to inspect the values of the image
            vmin = np.nanmin(data)
            vmax = np.nanmax(data)
        if black_bg == 'auto':
            # Guess if the background is rather black or light based on
            # the values of voxels near the border
            background = np.median(get_border_data(data, 2))
            if background > .5 * (vmin + vmax):
                black_bg = False
            else:
                black_bg = True
    if dim:
        if dim != 'auto' and not isinstance(dim, numbers.Number):
            raise ValueError(
                "The input given for 'dim' needs to be a float. "
                "You provided dim=%s in %s" % (str(dim), type(dim)))
        vmean = .5 * (vmin + vmax)
        ptp = .5 * (vmax - vmin)
        if black_bg:
            if not isinstance(dim, numbers.Number):
                dim = .8
            vmax = vmean + (1 + dim) * ptp
        else:
            if not isinstance(dim, numbers.Number):
                dim = .6
            vmin = .5 * (2 - dim) * vmean - (1 + dim) * ptp
    return anat_img, black_bg, vmin, vmax


###############################################################################
# Usage-specific functions


def plot_anat(anat_img=MNI152TEMPLATE, cut_coords=None,
              output_file=None, display_mode='ortho', figure=None,
              axes=None, title=None, annotate=True, threshold=None,
              draw_cross=True, black_bg='auto', dim='auto', cmap=plt.cm.gray,
              vmin=None, vmax=None, **kwargs):
    """Plot cuts of an anatomical image (by default 3 cuts:
    Frontal, Axial, and Lateral)

    Parameters
    ----------
    anat_img : Niimg-like object, optional
        See http://nilearn.github.io/manipulating_images/input_output.html
        The anatomical image to be used as a background. If None is
        given, nilearn tries to find a T1 template.
        Default=MNI152TEMPLATE.

    cut_coords : None, a tuple of floats, or an integer, optional
        The MNI coordinates of the point where the cut is performed
        If display_mode is 'ortho' or 'tiled', this should be a 3-tuple: (x, y, z)
        For display_mode == 'x', 'y', or 'z', then these are the
        coordinates of each cut in the corresponding direction.
        If None is given, the cuts is calculated automaticaly.
        If display_mode is 'x', 'y' or 'z', cut_coords can be an integer,
        in which case it specifies the number of cuts to perform.
        If display_mode is 'mosaic', and the number of cuts is the same
        for all directions, cut_coords can be specified as an integer.
        It can also be a length 3 tuple specifying the number of cuts for
        every direction if these are different.

    output_file : string, or None, optional
        The name of an image file to export the plot to. Valid extensions
        are .png, .pdf, .svg. If output_file is not None, the plot
        is saved to a file, and the display is closed.

    display_mode : {'ortho', 'tiled', 'mosaic', 'x', 'y', 'z', 'yx', 'xz', 'yz'}, optional
        Choose the direction of the cuts: 'x' - sagittal, 'y' - coronal,
        'z' - axial, 'ortho' - three cuts are performed in orthogonal
        directions, 'tiled' - three cuts are performed
        and arranged in a 2x2 grid, 'mosaic' - three cuts
        are performed along multiple rows and columns, Default='ortho'.

    figure : integer or matplotlib figure, optional
        Matplotlib figure used or its number. If None is given, a
        new figure is created.

    axes : matplotlib axes or 4 tuple of float: (xmin, ymin, width, height), optional
        The axes, or the coordinates, in matplotlib figure space,
        of the axes used to display the plot. If None, the complete figure is used.

    title : string, optional
        The title displayed on the figure.

    annotate : boolean, optional
        If annotate is True, positions and left/right annotation
        are added to the plot. Default=True.

    threshold : a number, None, or 'auto', optional
        If None is given, the image is not thresholded.
        If a number is given, it is used to threshold the image:
        values below the threshold (in absolute value) are plotted
        as transparent. If auto is given, the threshold is determined
        magically by analysis of the image.

    draw_cross : boolean, optional
        If draw_cross is True, a cross is drawn on the plot to
        indicate the cut plosition. Default=True.

    black_bg : boolean or 'auto', optional
        If True, the background of the image is set to be black. If
        you wish to save figures with a black background, you
        will need to pass "facecolor='k', edgecolor='k'"
        to matplotlib.pyplot.savefig.
        Default='auto'.

    dim : float or 'auto', optional
        Dimming factor applied to background image. By default, automatic
        heuristics are applied based upon the image intensity.
        Accepted float values, where a typical span is between -2 and 2
        (-2 = increase contrast; 2 = decrease contrast), but larger
        values can be used for a more pronounced effect. 0 means no
        dimming. Default='auto'.

    cmap : matplotlib colormap, optional
        The colormap for the anat. Default=plt.cm.gray.

    vmin : float, optional
        Lower bound for plotting, passed to matplotlib.pyplot.imshow.

    vmax : float, optional
        Upper bound for plotting, passed to matplotlib.pyplot.imshow.

    Notes
    -----
    Arrays should be passed in numpy convention: (x, y, z) ordered.

    For visualization, non-finite values found in passed 'anat_img'
    are set to zero.

    """
    anat_img, black_bg, anat_vmin, anat_vmax = _load_anat(
        anat_img,
        dim=dim, black_bg=black_bg)

    if vmin is None:
        vmin = anat_vmin
    if vmax is None:
        vmax = anat_vmax

    display = plot_img(anat_img, cut_coords=cut_coords,
                       output_file=output_file, display_mode=display_mode,
                       figure=figure, axes=axes, title=title,
                       threshold=threshold, annotate=annotate,
                       draw_cross=draw_cross, black_bg=black_bg,
                       vmin=vmin, vmax=vmax, cmap=cmap, **kwargs)
    return display


def plot_epi(epi_img=None, cut_coords=None, output_file=None,
             display_mode='ortho', figure=None, axes=None, title=None,
             annotate=True, draw_cross=True, black_bg=True,
             cmap=plt.cm.nipy_spectral, vmin=None, vmax=None, **kwargs):
    """Plot cuts of an EPI image (by default 3 cuts:
    Frontal, Axial, and Lateral)

    Parameters
    ----------
    epi_img : a nifti-image like object or a filename, optional
        The EPI (T2*) image.

    cut_coords : None, a tuple of floats, or an integer, optional
        The MNI coordinates of the point where the cut is performed
        If display_mode is 'ortho' or 'tiled', this should be a 3-tuple: (x, y, z)
        For display_mode == 'x', 'y', or 'z', then these are the
        coordinates of each cut in the corresponding direction.
        If None is given, the cuts is calculated automaticaly.
        If display_mode is 'x', 'y' or 'z', cut_coords can be an integer,
        in which case it specifies the number of cuts to perform.
        If display_mode is 'mosaic', and the number of cuts is the same
        for all directions, cut_coords can be specified as an integer.
        It can also be a length 3 tuple specifying the number of cuts for
        every direction if these are different.

    output_file : string, or None, optional
        The name of an image file to export the plot to. Valid extensions
        are .png, .pdf, .svg. If output_file is not None, the plot
        is saved to a file, and the display is closed.

    display_mode : {'ortho', 'tiled', 'mosaic', 'x', 'y', 'z', 'yx', 'xz', 'yz'}, optional
        Choose the direction of the cuts: 'x' - sagittal, 'y' - coronal,
        'z' - axial, 'ortho' - three cuts are performed in orthogonal
        directions, 'tiled' - three cuts are performed
        and arranged in a 2x2 grid. 'mosaic' - three cuts are
        performed along multiple rows and columns. Default='ortho'.

    figure : integer or matplotlib figure, optional
        Matplotlib figure used or its number. If None is given, a
        new figure is created.

    axes : matplotlib axes or 4 tuple of float: (xmin, ymin, width, height), optional
        The axes, or the coordinates, in matplotlib figure space,
        of the axes used to display the plot. If None, the complete figure is used.

    title : string, optional
        The title displayed on the figure.

    annotate : boolean, optional
        If annotate is True, positions and left/right annotation
        are added to the plot. Default=True.

    draw_cross : boolean, optional
        If draw_cross is True, a cross is drawn on the plot to
        indicate the cut plosition. Default=True.

    black_bg : boolean, optional
        If True, the background of the image is set to be black. If
        you wish to save figures with a black background, you
        will need to pass "facecolor='k', edgecolor='k'"
        to matplotlib.pyplot.savefig.
        Default=True.

    cmap : matplotlib colormap, optional
        The colormap for specified image.
        Default=plt.cm.nipy_spectral.

    threshold : a number, None, or 'auto', optional
        If None is given, the image is not thresholded.
        If a number is given, it is used to threshold the image:
        values below the threshold (in absolute value) are plotted
        as transparent. If auto is given, the threshold is determined
        magically by analysis of the image.

    vmin : float, optional
        Lower bound for plotting, passed to matplotlib.pyplot.imshow.

    vmax : float, optional
        Upper bound for plotting, passed to matplotlib.pyplot.imshow.

    Notes
    -----
    Arrays should be passed in numpy convention: (x, y, z) ordered.

    """
    display = plot_img(epi_img, cut_coords=cut_coords,
                       output_file=output_file, display_mode=display_mode,
                       figure=figure, axes=axes, title=title,
                       threshold=None, annotate=annotate,
                       draw_cross=draw_cross, black_bg=black_bg,
                       cmap=cmap, vmin=vmin, vmax=vmax, **kwargs)
    return display


def _plot_roi_contours(display, roi_img, cmap, alpha, linewidths):
    """Helper function for plotting regions of interest ROIs in contours.

    Parameters
    ----------
    display : nilearn.plotting.displays.OrthoSlicer, object
        An object with background image on which contours are shown.

    roi_img : Niimg-like object
        See http://nilearn.github.io/manipulating_images/input_output.html
        The ROI/mask image, it could be binary mask or an atlas or ROIs
        with integer values.

    cmap : matplotlib colormap
        The colormap for the atlas maps.

    alpha : float between 0 and 1
        Alpha sets the transparency of the color inside the filled
        contours.

    linewidths : float
        This option can be used to set the boundary thickness of the
        contours.

    Returns
    -------
    display : nilearn.plotting.displays.OrthoSlicer, object
        Contours displayed on the background image.

    """
    roi_img = _utils.check_niimg_3d(roi_img)
    roi_data = get_data(roi_img)
    labels = np.unique(roi_data)
    cmap = plt.cm.get_cmap(cmap)
    color_list = cmap(np.linspace(0, 1, len(labels)))
    for idx, label in enumerate(labels):
        if label == 0:
            continue
        data = (roi_data == label)
        data = data.astype(int)
        img = new_img_like(roi_img, data, affine=roi_img.affine)
        display.add_contours(img, levels=[0.5], colors=[color_list[idx - 1]],
                             alpha=alpha, linewidths=linewidths,
                             linestyles='solid')
    return display


def plot_roi(roi_img, bg_img=MNI152TEMPLATE, cut_coords=None,
             output_file=None, display_mode='ortho', figure=None, axes=None,
             title=None, annotate=True, draw_cross=True, black_bg='auto',
             threshold=0.5, alpha=0.7, cmap=plt.cm.gist_ncar, dim='auto',
             vmin=None, vmax=None, resampling_interpolation='nearest',
             view_type='continuous', linewidths=2.5, **kwargs):
    """Plot cuts of an ROI/mask image (by default 3 cuts: Frontal, Axial, and
    Lateral)

    Parameters
    ----------
    roi_img : Niimg-like object
        See http://nilearn.github.io/manipulating_images/input_output.html
        The ROI/mask image, it could be binary mask or an atlas or ROIs
        with integer values.

    bg_img : Niimg-like object, optional
        See http://nilearn.github.io/manipulating_images/input_output.html
        The background image that the ROI/mask will be plotted on top of.
        If nothing is specified, the MNI152 template will be used.
        To turn off background image, just pass "bg_img=None".
        Default=MNI152TEMPLATE.

    cut_coords : None, or a tuple of floats, optional
        The MNI coordinates of the point where the cut is performed, in
        MNI coordinates and order.
        If display_mode is 'ortho' or 'tiled', this should be a 3-tuple: (x, y, z)
        For display_mode == 'x', 'y', or 'z', then these are the
        coordinates of each cut in the corresponding direction.
        If None is given, the cuts is calculated automaticaly.
        If display_mode is 'mosaic', and the number of cuts is the same
        for all directions, cut_coords can be specified as an integer.
        It can also be a length 3 tuple specifying the number of cuts for
        every direction if these are different.

    output_file : string, or None, optional
        The name of an image file to export the plot to. Valid extensions
        are .png, .pdf, .svg. If output_file is not None, the plot
        is saved to a file, and the display is closed.

    display_mode : {'ortho', 'mosaic', 'tiled', 'x', 'y', 'z', 'yx', 'xz', 'yz'}, optional
        Choose the direction of the cuts: 'x' - sagittal, 'y' - coronal,
        'z' - axial, 'ortho' - three cuts are performed in orthogonal
        directions, 'tiled' - three cuts are performed
        and arranged in a 2x2 grid, 'mosaic' - three cuts
        are performed along multiple rows and columns, Default='ortho'.

    figure : integer or matplotlib figure, optional
        Matplotlib figure used or its number. If None is given, a
        new figure is created.

    axes : matplotlib axes or 4 tuple of float: (xmin, ymin, width, height), optional
        The axes, or the coordinates, in matplotlib figure space,
        of the axes used to display the plot. If None, the complete
        figure is used.

    title : string, optional
        The title displayed on the figure.

    annotate : boolean, optional
        If annotate is True, positions and left/right annotation
        are added to the plot. Default=True.

    draw_cross : boolean, optional
        If draw_cross is True, a cross is drawn on the plot to
        indicate the cut plosition.

    black_bg : boolean or 'auto', optional
        If True, the background of the image is set to be black. If
        you wish to save figures with a black background, you
        will need to pass "facecolor='k', edgecolor='k'"
        to matplotlib.pyplot.savefig.
        Default='auto'.

    threshold : None, 'auto', or a number, optional
        If None is given, the image is not thresholded.
        If a number is given, it is used to threshold the image:
        values below the threshold (in absolute value) are plotted
        as transparent. If auto is given, the threshold is determined
        magically by analysis of the image.
        Default=0.5.

    alpha : float between 0 and 1, optional
        Alpha sets the transparency of the color inside the filled
        contours. Default=0.7.

    cmap : matplotlib colormap, optional
        The colormap for the atlas maps. Default=plt.cm.gist_ncar.

    dim : float or 'auto', optional
        Dimming factor applied to background image. By default, automatic
        heuristics are applied based upon the background image intensity.
        Accepted float values, where a typical span is between -2 and 2
        (-2 = increase contrast; 2 = decrease contrast), but larger values
        can be used for a more pronounced effect. 0 means no dimming.
        Default='auto'.

    vmin : float, optional
        Lower bound for plotting, passed to matplotlib.pyplot.imshow.

    vmax : float, optional
        Upper bound for plotting, passed to matplotlib.pyplot.imshow.

    resampling_interpolation : str, optional
        Interpolation to use when resampling the image to the destination
        space. Can be "continuous" to use 3rd-order spline interpolation,
        or "nearest" to use nearest-neighbor mapping.
        "nearest" is faster but can be noisier in some cases.
        Default='nearest'.

    view_type : {'continuous', 'contours'}, optional
        By default view_type == 'continuous', rois are shown as continuous colors.
        If view_type == 'contours', maps are shown as contours. For this type, label
        denoted as 0 is considered as background and not shown.
        Default='continuous'.

    linewidths : float, optional
        This option can be used to set the boundary thickness of the
        contours. Only reflects when view_type='contours'.
        Default=2.5.

    Notes
    -----
    A small threshold is applied by default to eliminate numerical
    background noise.

    For visualization, non-finite values found in passed 'roi_img' or
    'bg_img' are set to zero.

    See Also
    --------
    nilearn.plotting.plot_prob_atlas : To simply plot probabilistic atlases
        (4D images)

    """  # noqa: E501
    valid_view_types = ['continuous', 'contours']
    if view_type not in valid_view_types:
        raise ValueError(
            'Unknown view type: %s. Valid view types are %s' %
            (str(view_type), str(valid_view_types))
        )
    elif view_type == 'contours':
        img = roi_img
        roi_img = None

    bg_img, black_bg, bg_vmin, bg_vmax = _load_anat(bg_img, dim=dim,
                                                    black_bg=black_bg)

    display = _plot_img_with_bg(
        img=roi_img, bg_img=bg_img, cut_coords=cut_coords,
        output_file=output_file, display_mode=display_mode,
        figure=figure, axes=axes, title=title, annotate=annotate,
        draw_cross=draw_cross, black_bg=black_bg,
        threshold=threshold, bg_vmin=bg_vmin, bg_vmax=bg_vmax,
        resampling_interpolation=resampling_interpolation,
        alpha=alpha, cmap=cmap, vmin=vmin, vmax=vmax, **kwargs)

    if view_type == 'contours':
        display = _plot_roi_contours(display, img, cmap=cmap, alpha=alpha,
                                     linewidths=linewidths)

    return display


def plot_prob_atlas(maps_img, bg_img=MNI152TEMPLATE, view_type='auto',
                    threshold='auto', linewidths=2.5, cut_coords=None,
                    output_file=None, display_mode='ortho',
                    figure=None, axes=None, title=None, annotate=True,
                    draw_cross=True, black_bg='auto', dim='auto',
                    colorbar=False,
                    cmap=plt.cm.gist_rainbow, vmin=None, vmax=None,
                    alpha=0.7, **kwargs):
    """Plot the probabilistic atlases onto the anatomical image
    by default MNI template

    Parameters
    ----------
    maps_img : Niimg-like object or the filename
        4D image of the probabilistic atlas maps.

    bg_img : Niimg-like object, optional
        See http://nilearn.github.io/manipulating_images/input_output.html
        The anatomical image to be used as a background.
        If nothing is specified, the MNI152 template will be used.
        To turn off background image, just pass "bg_img=False".
        Default=MNI152TEMPLATE.

        .. versionadded:: 0.4.0

    view_type : {'auto', 'contours', 'filled_contours', 'continuous'}, optional
        By default view_type == 'auto', means maps will be displayed
        automatically using any one of the three view types. The automatic
        selection of view type depends on the total number of maps.
        If view_type == 'contours', maps are overlayed as contours
        If view_type == 'filled_contours', maps are overlayed as contours
        along with color fillings inside the contours.
        If view_type == 'continuous', maps are overlayed as continous
        colors irrespective of the number maps.
        Default='auto'.

    threshold : a str or a number, list of str or numbers, optional
        This parameter is optional and is used to threshold the maps image
        using the given value or automatically selected value. The values
        in the image above the threshold level will be visualized.
        The default strategy, computes a threshold level that seeks to
        minimize (yet not eliminate completely) the overlap between several
        maps for a better visualization.
        The threshold can also be expressed as a percentile over the values
        of the whole atlas. In that case, the value must be specified as
        string finishing with a percent sign, e.g., "25.3%".
        If a single string is provided, the same percentile will be applied
        over the whole atlas. Otherwise, if a list of percentiles is
        provided, each 3D map is thresholded with certain percentile
        sequentially. Length of percentiles given should match the number
        of 3D map in time (4th) dimension.
        If a number or a list of numbers, the given value will be used
        directly to threshold the maps without any percentile calculation.
        If None, a very small threshold is applied to remove numerical
        noise from the maps background.

    linewidths : float, optional
        This option can be used to set the boundary thickness of the contours.
        Default=2.5.

    cut_coords : None, a tuple of floats, or an integer, optional
        The MNI coordinates of the point where the cut is performed
        If display_mode is 'ortho' or 'tiled', this should be a 3-tuple: (x, y, z)
        For display_mode == 'x', 'y', or 'z', then these are the
        coordinates of each cut in the corresponding direction.
        If None is given, the cuts is calculated automaticaly.
        If display_mode is 'x', 'y' or 'z', cut_coords can be an integer,
        in which case it specifies the number of cuts to perform.
        If display_mode is 'mosaic', and the number of cuts is the same
        for all directions, cut_coords can be specified as an integer.
        It can also be a length 3 tuple specifying the number of cuts for
        every direction if these are different.

    output_file : string, or None, optional
        The name of an image file to export the plot to. Valid extensions
        are .png, .pdf, .svg. If output_file is not None, the plot
        is saved to a file, and the display is closed.

    display_mode : {'ortho', 'tiled', 'mosaic', 'x', 'y', 'z', 'yx', 'xz', 'yz'}, optional
        Choose the direction of the cuts: 'x' - sagittal, 'y' - coronal,
        'z' - axial, 'ortho' - three cuts are performed in orthogonal
        directions, 'tiled' - three cuts are performed
        and arranged in a 2x2 grid, 'mosaic' - three cuts are performed along
        multiple rows and columns, Default='ortho'.

    figure : integer or matplotlib figure, optional
        Matplotlib figure used or its number. If None is given, a
        new figure is created.

    axes : matplotlib axes or 4 tuple of float: (xmin, ymin, width, height), optional
        The axes, or the coordinates, in matplotlib figure space,
        of the axes used to display the plot. If None, the complete figure is used.

    title : string, optional
        The title displayed on the figure.

    annotate : boolean, optional
        If annotate is True, positions and left/right annotation
        are added to the plot. Default=True.

    draw_cross : boolean, optional
        If draw_cross is True, a cross is drawn on the plot to
        indicate the cut plosition. Default=True.

    black_bg : boolean or 'auto', optional
        If True, the background of the image is set to be black. If
        you wish to save figures with a black background, you
        will need to pass "facecolor='k', edgecolor='k'" to pylab's
        savefig. Default='auto'.

    dim : float or 'auto', optional
        Dimming factor applied to background image. By default, automatic
        heuristics are applied based upon the background image intensity.
        Accepted float values, where a typical span is between -2 and 2
        (-2 = increase contrast; 2 = decrease contrast), but larger values
        can be used for a more pronounced effect. 0 means no dimming.
        Default='auto'.

    cmap : matplotlib colormap, optional
        The colormap for the atlas maps. Default=plt.cm.gist_rainbow.

    colorbar : boolean, optional
        If True, display a colorbar on the right of the plots.
        Default=False.

    vmin : float, optional
        Lower bound for plotting, passed to matplotlib.pyplot.imshow.

    vmax : float, optional
        Upper bound for plotting, passed to matplotlib.pyplot.imshow.

    alpha : float between 0 and 1, optional
        Alpha sets the transparency of the color inside the filled contours.
        Default=0.7.

    See Also
    --------
    nilearn.plotting.plot_roi : To simply plot max-prob atlases (3D images)

    """
    display = plot_anat(bg_img, cut_coords=cut_coords,
                        display_mode=display_mode,
                        figure=figure, axes=axes, title=title,
                        annotate=annotate, draw_cross=draw_cross,
                        black_bg=black_bg, dim=dim, **kwargs)

    maps_img = _utils.check_niimg_4d(maps_img)
    n_maps = maps_img.shape[3]

    valid_view_types = ['auto', 'contours', 'filled_contours', 'continuous']
    if view_type not in valid_view_types:
        raise ValueError(
            'Unknown view type: %s. Valid view types are %s' %
            (str(view_type), str(valid_view_types))
        )

    cmap = plt.cm.get_cmap(cmap)
    color_list = cmap(np.linspace(0, 1, n_maps))

    if view_type == 'auto':
        if n_maps > 20:
            view_type = 'contours'
        elif n_maps > 10:
            view_type = 'filled_contours'
        else:
            view_type = 'continuous'

    if threshold is None:
        threshold = 1e-6
    elif threshold == 'auto':
        # it will use default percentage,
        # strategy is to avoid maximum overlaps as possible
        if view_type == 'contours':
            correction_factor = 1
        elif view_type == 'filled_contours':
            correction_factor = .8
        else:
            correction_factor = .5
        threshold = "%f%%" % (100 * (1 - .2 * correction_factor / n_maps))

    if (isinstance(threshold, collections.abc.Iterable) and
            not isinstance(threshold, str)):
        threshold = [thr for thr in threshold]
        if len(threshold) != n_maps:
            raise TypeError('The list of values to threshold '
                            'should be equal to number of maps')
    else:
        threshold = [threshold] * n_maps

    filled = view_type.startswith('filled')
    for (map_img, color, thr) in zip(iter_img(maps_img), color_list,
                                     threshold):
        data = get_data(map_img)
        # To threshold or choose the level of the contours
        thr = check_threshold(thr, data,
                              percentile_func=fast_abs_percentile,
                              name='threshold')
        # Get rid of background values in all cases
        thr = max(thr, 1e-6)

        if view_type == 'continuous':
            display.add_overlay(map_img, threshold=thr,
                                cmap=cm.alpha_cmap(color), alpha=alpha)
        else:
            display.add_contours(map_img, levels=[thr],
                                 linewidths=linewidths,
                                 colors=[color], filled=filled,
                                 alpha=alpha, linestyles='solid', )
    if colorbar:
        display._colorbar = True
        # Create a colormap from color list to feed display
        cmap = matplotlib.colors.LinearSegmentedColormap.from_list(
            'segmented colors', color_list, n_maps + 1)
        display._show_colorbar(cmap, matplotlib.colors.Normalize(1,
                                                                 n_maps + 1))
        tick_locator = matplotlib.ticker.MaxNLocator(nbins=10)
        display.locator = tick_locator
        display._cbar.update_ticks()
        tick_location = np.round(np.linspace(1,
                                             n_maps,
                                             min(n_maps, 10))).astype('int')
        display._cbar.set_ticks(tick_location + .5)
        display._cbar.set_ticklabels(tick_location)
        left, bottom, width, height = display._colorbar_ax.\
            get_position().bounds
        display._colorbar_ax.set_position([left, bottom, width, height * 0.95])
        display._colorbar_ax.annotate('Map #', xy=(1, 1.03), ha='right',
                                      va='bottom',
                                      xycoords='axes fraction')
    if output_file is not None:
        display.savefig(output_file)
        display.close()
        display = None

    return display


def plot_stat_map(stat_map_img, bg_img=MNI152TEMPLATE, cut_coords=None,
                  output_file=None, display_mode='ortho', colorbar=True,
                  figure=None, axes=None, title=None, threshold=1e-6,
                  annotate=True, draw_cross=True, black_bg='auto',
                  cmap=cm.cold_hot, symmetric_cbar="auto",
                  dim='auto', vmax=None, resampling_interpolation='continuous',
                  **kwargs):
    """Plot cuts of an ROI/mask image (by default 3 cuts: Frontal, Axial, and
    Lateral)

    Parameters
    ----------
    stat_map_img : Niimg-like object
        See http://nilearn.github.io/manipulating_images/input_output.html
        The statistical map image

    bg_img : Niimg-like object, optional
        See http://nilearn.github.io/manipulating_images/input_output.html
        The background image that the ROI/mask will be plotted on top of.
        If nothing is specified, the MNI152 template will be used.
        To turn off background image, just pass "bg_img=None".
        Default=MNI152TEMPLATE.

    cut_coords : None, a tuple of floats, or an integer, optional
        The MNI coordinates of the point where the cut is performed
        If display_mode is 'ortho' or 'tiled', this should be a 3-tuple: (x, y, z)
        For display_mode == 'x', 'y', or 'z', then these are the
        coordinates of each cut in the corresponding direction.
        If None is given, the cuts is calculated automaticaly.
        If display_mode is 'x', 'y' or 'z', cut_coords can be an integer,
        in which case it specifies the number of cuts to perform.
        If display_mode is 'mosaic', and the number of cuts is the same
        for all directions, cut_coords can be specified as an integer.
        It can also be a length 3 tuple specifying the number of cuts for
        every direction if these are different.

    output_file : string, or None, optional
        The name of an image file to export the plot to. Valid extensions
        are .png, .pdf, .svg. If output_file is not None, the plot
        is saved to a file, and the display is closed.

    display_mode : {'ortho', 'tiled', 'mosaic', 'x', 'y', 'z', 'yx', 'xz', 'yz'}, optional
        Choose the direction of the cuts: 'x' - sagittal, 'y' - coronal,
        'z' - axial, 'ortho' - three cuts are performed in orthogonal
        directions, 'tiled' - three cuts are performed
        and arranged in a 2x2 grid, 'mosaic' - three cuts
        are performed along multiple rows and columns, Default='ortho'.

    colorbar : boolean, optional
        If True, display a colorbar on the right of the plots. Default=True.

    figure : integer or matplotlib figure, optional
        Matplotlib figure used or its number. If None is given, a
        new figure is created.

    axes : matplotlib axes or 4 tuple of float: (xmin, ymin, width, height), optional
        The axes, or the coordinates, in matplotlib figure space,
        of the axes used to display the plot. If None, the complete figure is used.

    title : string, optional
        The title displayed on the figure.

    threshold : a number, None, or 'auto', optional
        If None is given, the image is not thresholded.
        If a number is given, it is used to threshold the image:
        values below the threshold (in absolute value) are plotted
        as transparent. If auto is given, the threshold is determined
        magically by analysis of the image.
        Default=1e-6.

    annotate : boolean, optional
        If annotate is True, positions and left/right annotation
        are added to the plot. Default=True.

    draw_cross : boolean, optional
        If draw_cross is True, a cross is drawn on the plot to
        indicate the cut plosition. Default=True.

    black_bg : boolean or 'auto', optional
        If True, the background of the image is set to be black. If
        you wish to save figures with a black background, you
        will need to pass "facecolor='k', edgecolor='k'"
        to matplotlib.pyplot.savefig.
        Default='auto'.

    cmap : matplotlib colormap, optional
        The colormap for specified image. The ccolormap *must* be symmetrical.
        Default=plt.cm.cold_hot.

    symmetric_cbar : boolean or 'auto', optional
        Specifies whether the colorbar should range from -vmax to vmax
        or from vmin to vmax. Setting to 'auto' will select the latter if
        the range of the whole image is either positive or negative.
        Note: The colormap will always be set to range from -vmax to vmax.
        Default='auto'.

    dim : float or 'auto', optional
        Dimming factor applied to background image. By default, automatic
        heuristics are applied based upon the background image intensity.
        Accepted float values, where a typical scan is between -2 and 2
        (-2 = increase constrast; 2 = decrease contrast), but larger values
        can be used for a more pronounced effect. 0 means no dimming.
        Default='auto'.

    vmax : float, optional
        Upper bound for plotting, passed to matplotlib.pyplot.imshow.

    resampling_interpolation : str, optional
        Interpolation to use when resampling the image to the destination
        space. Can be "continuous" to use 3rd-order spline interpolation,
        or "nearest" to use nearest-neighbor mapping.
        "nearest" is faster but can be noisier in some cases.
        Default='continuous'.

    Notes
    -----
    Arrays should be passed in numpy convention: (x, y, z) ordered.

    For visualization, non-finite values found in passed 'stat_map_img' or
    'bg_img' are set to zero.

    See Also
    --------
    nilearn.plotting.plot_anat : To simply plot anatomical images
    nilearn.plotting.plot_epi : To simply plot raw EPI images
    nilearn.plotting.plot_glass_brain : To plot maps in a glass brain

    """  # noqa: E501
    # dim the background
    bg_img, black_bg, bg_vmin, bg_vmax = _load_anat(bg_img, dim=dim,
                                                    black_bg=black_bg)

    stat_map_img = _utils.check_niimg_3d(stat_map_img, dtype='auto')

    cbar_vmin, cbar_vmax, vmin, vmax = _get_colorbar_and_data_ranges(
        _safe_get_data(stat_map_img, ensure_finite=True),
        vmax,
        symmetric_cbar,
        kwargs)

    display = _plot_img_with_bg(
        img=stat_map_img, bg_img=bg_img, cut_coords=cut_coords,
        output_file=output_file, display_mode=display_mode,
        figure=figure, axes=axes, title=title, annotate=annotate,
        draw_cross=draw_cross, black_bg=black_bg, threshold=threshold,
        bg_vmin=bg_vmin, bg_vmax=bg_vmax, cmap=cmap, vmin=vmin, vmax=vmax,
        colorbar=colorbar, cbar_vmin=cbar_vmin, cbar_vmax=cbar_vmax,
        resampling_interpolation=resampling_interpolation, **kwargs)

    return display


def plot_glass_brain(stat_map_img,
                     output_file=None, display_mode='ortho', colorbar=False,
                     figure=None, axes=None, title=None, threshold='auto',
                     annotate=True,
                     black_bg=False,
                     cmap=None,
                     alpha=0.7,
                     vmin=None, vmax=None,
                     plot_abs=True,
                     symmetric_cbar="auto",
                     resampling_interpolation='continuous',
                     **kwargs):
    """Plot 2d projections of an ROI/mask image (by default 3 projections:
    Frontal, Axial, and Lateral). The brain glass schematics
    are added on top of the image.

    The plotted image should be in MNI space for this function to work
    properly.

    Only glass brain can be plotted by switching stat_map_img to None.

    Parameters
    ----------
    stat_map_img : Niimg-like object
        See http://nilearn.github.io/manipulating_images/input_output.html
        The statistical map image. It needs to be in MNI space
        in order to align with the brain schematics.

    output_file : string, or None, optional
        The name of an image file to export the plot to. Valid extensions
        are .png, .pdf, .svg. If output_file is not None, the plot
        is saved to a file, and the display is closed.

    display_mode : string, optional
        Choose the direction of the cuts: 'x' - sagittal, 'y' - coronal,
        'z' - axial, 'l' - sagittal left hemisphere only,
        'r' - sagittal right hemisphere only, 'ortho' - three cuts are
        performed in orthogonal directions. Possible values are: 'ortho',
        'x', 'y', 'z', 'xz', 'yx', 'yz', 'l', 'r', 'lr', 'lzr', 'lyr',
        'lzry', 'lyrz'. Default='ortho'.

    colorbar : boolean, optional
        If True, display a colorbar on the right of the plots.
        Default=False.

    figure : integer or matplotlib figure, optional
        Matplotlib figure used or its number. If None is given, a
        new figure is created.

    axes : matplotlib axes or 4 tuple of float: (xmin, ymin, width, height), optional
        The axes, or the coordinates, in matplotlib figure space,
        of the axes used to display the plot. If None, the complete figure is used.

    title : string, optional
        The title displayed on the figure.

    threshold : a number or 'auto', optional
            If None is given, the image is not thresholded.
            If a number is given, it is used to threshold the image:
            values below the threshold (in absolute value) are plotted
            as transparent. If auto is given, the threshold is determined
            magically by analysis of the image.
            Default='auto'.

    annotate : boolean, optional
        If annotate is True, positions and left/right annotation
        are added to the plot. Default=True.

    black_bg : boolean, optional
        If True, the background of the image is set to be black. If
        you wish to save figures with a black background, you
        will need to pass "facecolor='k', edgecolor='k'"
        to matplotlib.pyplot.savefig. Default=False.

    cmap : matplotlib colormap, optional
        The colormap for specified image

    alpha : float between 0 and 1, optional
        Alpha transparency for the brain schematics. Default=0.7.

    vmin : float, optional
        Lower bound for plotting, passed to matplotlib.pyplot.imshow.

    vmax : float, optional
        Upper bound for plotting, passed to matplotlib.pyplot.imshow.

    plot_abs : boolean, optional
        If set to True (default) maximum intensity projection of the
        absolute value will be used (rendering positive and negative
        values in the same manner). If set to false the sign of the
        maximum intensity will be represented with different colors.
        See http://nilearn.github.io/auto_examples/01_plotting/plot_demo_glass_brain_extensive.html
        for examples. Default=True.

    symmetric_cbar : boolean or 'auto', optional
        Specifies whether the colorbar should range from -vmax to vmax
        or from vmin to vmax. Setting to 'auto' will select the latter if
        the range of the whole image is either positive or negative.
        Note: The colormap will always be set to range from -vmax to vmax.
        Default='auto'.

    resampling_interpolation : str, optional
        Interpolation to use when resampling the image to the destination
        space. Can be "continuous" (default) to use 3rd-order spline
        interpolation, or "nearest" to use nearest-neighbor mapping.
        "nearest" is faster but can be noisier in some cases.
        Default='continuous'.

    Notes
    -----
    Arrays should be passed in numpy convention: (x, y, z) ordered.

    """
    if cmap is None:
        cmap = cm.cold_hot if black_bg else cm.cold_white_hot

    if stat_map_img:
        stat_map_img = _utils.check_niimg_3d(stat_map_img, dtype='auto')
        if plot_abs:
            cbar_vmin, cbar_vmax, vmin, vmax = _get_colorbar_and_data_ranges(
                _safe_get_data(stat_map_img, ensure_finite=True),
                vmax,
                symmetric_cbar,
                kwargs,
                0)
        else:
            cbar_vmin, cbar_vmax, vmin, vmax = _get_colorbar_and_data_ranges(
                _safe_get_data(stat_map_img, ensure_finite=True),
                vmax,
                symmetric_cbar,
                kwargs)
    else:
        cbar_vmin, cbar_vmax = None, None

    def display_factory(display_mode):
        return functools.partial(get_projector(display_mode),
                                 alpha=alpha, plot_abs=plot_abs)

    display = _plot_img_with_bg(
        img=stat_map_img, output_file=output_file, display_mode=display_mode,
        figure=figure, axes=axes, title=title, annotate=annotate,
        black_bg=black_bg, threshold=threshold, cmap=cmap, colorbar=colorbar,
        display_factory=display_factory, vmin=vmin, vmax=vmax,
        cbar_vmin=cbar_vmin, cbar_vmax=cbar_vmax,
        resampling_interpolation=resampling_interpolation, **kwargs)

    if stat_map_img is None and 'l' in display.axes:
        display.axes['l'].ax.invert_xaxis()

    return display


def plot_connectome(adjacency_matrix, node_coords,
                    node_color='auto', node_size=50,
                    edge_cmap=cm.bwr,
                    edge_vmin=None, edge_vmax=None,
                    edge_threshold=None,
                    output_file=None, display_mode='ortho',
                    figure=None, axes=None, title=None,
                    annotate=True, black_bg=False,
                    alpha=0.7,
                    edge_kwargs=None, node_kwargs=None,
                    colorbar=False):
    """Plot connectome on top of the brain glass schematics.

    The plotted image should be in MNI space for this function to work
    properly.

    In the case of 'l' and 'r' directions (for hemispheric projections),
    markers in the coordinate x == 0 are included in both hemispheres.

    Parameters
    ----------
    adjacency_matrix : numpy array of shape (n, n)
        Represents the link strengths of the graph. The matrix can be
        symmetric which will result in an undirected graph, or not
        symmetric which will result in a directed graph.

    node_coords : numpy array_like of shape (n, 3)
        3d coordinates of the graph nodes in world space.

    node_color : color or sequence of colors or 'auto', optional
        Color(s) of the nodes. If string is given, all nodes
        are plotted with same color given in string.

    node_size : scalar or array_like, optional
        Size(s) of the nodes in points^2. Default=50.

    edge_cmap : colormap, optional
        Colormap used for representing the strength of the edges.
        Default=cm.bwr.

    edge_vmin, edge_vmax : float, optional
        If not None, either or both of these values will be used to
        as the minimum and maximum values to color edges. If None are
        supplied the maximum absolute value within the given threshold
        will be used as minimum (multiplied by -1) and maximum
        coloring levels.

    edge_threshold : str or number, optional
        If it is a number only the edges with a value greater than
        edge_threshold will be shown.
        If it is a string it must finish with a percent sign,
        e.g. "25.3%", and only the edges with a abs(value) above
        the given percentile will be shown.

    output_file : string, or None, optional
        The name of an image file to export the plot to. Valid extensions
        are .png, .pdf, .svg. If output_file is not None, the plot
        is saved to a file, and the display is closed.

    display_mode : string, optional
        Choose the direction of the cuts: 'x' - sagittal, 'y' - coronal,
        'z' - axial, 'l' - sagittal left hemisphere only,
        'r' - sagittal right hemisphere only, 'ortho' - three cuts are
        performed in orthogonal directions. Possible values are: 'ortho',
        'x', 'y', 'z', 'xz', 'yx', 'yz', 'l', 'r', 'lr', 'lzr', 'lyr',
        'lzry', 'lyrz'. Default='ortho'.

    figure : integer or matplotlib figure, optional
        Matplotlib figure used or its number.
        If None is given, a new figure is created.

    axes : matplotlib axes or 4 tuple of float: (xmin, ymin, width, height), optional
        The axes, or the coordinates, in matplotlib figure space,
        of the axes used to display the plot. If None, the complete figure is used.

    title : string, optional
        The title displayed on the figure.

    annotate : boolean, optional
        If annotate is True, positions and left/right annotation are
        added to the plot. Default=True.

    black_bg : boolean, optional
        If True, the background of the image is set to be black. If
        you wish to save figures with a black background, you
        will need to pass "facecolor='k', edgecolor='k'"
        to matplotlib.pyplot.savefig. Default=False.

    alpha : float between 0 and 1, optional
        Alpha transparency for the brain schematics. Default=0.7.

    edge_kwargs : dict, optional
        Will be passed as kwargs for each edge matlotlib Line2D.

    node_kwargs : dict, optional
        Will be passed as kwargs to the plt.scatter call that plots all
        the nodes in one go.

    colorbar : bool, optional
        If True, display a colorbar on the right of the plots.
        Default=False.

    See Also
    ---------
    nilearn.plotting.find_parcellation_cut_coords : Extraction of node
        coords on brain parcellations.
    nilearn.plotting.find_probabilistic_atlas_cut_coords : Extraction of
        node coords on brain probabilisitic atlases.

    """
    display = plot_glass_brain(None,
                               display_mode=display_mode,
                               figure=figure, axes=axes, title=title,
                               annotate=annotate,
                               black_bg=black_bg)

    display.add_graph(adjacency_matrix, node_coords,
                      node_color=node_color, node_size=node_size,
                      edge_cmap=edge_cmap,
                      edge_vmin=edge_vmin, edge_vmax=edge_vmax,
                      edge_threshold=edge_threshold,
                      edge_kwargs=edge_kwargs, node_kwargs=node_kwargs,
                      colorbar=colorbar)

    if output_file is not None:
        display.savefig(output_file)
        display.close()
        display = None

    return display


def plot_connectome_strength(adjacency_matrix, node_coords, node_size="auto",
                             cmap=None, output_file=None, display_mode="ortho",
                             figure=None, axes=None, title=None):
    """Plot connectome strength on top of the brain glass schematics.

    The strength of a connection is define as the sum of absolute values of
    the edges arriving to a node.

    Parameters
    ----------
    adjacency_matrix : numpy array of shape (n, n)
        Represents the link strengths of the graph. Assumed to be
        a symmetric matrix.

    node_coords : numpy array_like of shape (n, 3)
        3d coordinates of the graph nodes in world space.

    node_size : 'auto' or scalar, optional
        Size(s) of the nodes in points^2. By default the size of the node is
        inversely propertionnal to the number of nodes.
        Default='auto'.

    cmap : str or colormap, optional
        Colormap used to represent the strength of a node.

    output_file : string, or None, optional
        The name of an image file to export the plot to. Valid extensions
        are .png, .pdf, .svg. If output_file is not None, the plot
        is saved to a file, and the display is closed.

    display_mode : string, optional
        Choose the direction of the cuts: 'x' - sagittal, 'y' - coronal,
        'z' - axial, 'l' - sagittal left hemisphere only,
        'r' - sagittal right hemisphere only, 'ortho' - three cuts are
        performed in orthogonal directions. Possible values are: 'ortho',
        'x', 'y', 'z', 'xz', 'yx', 'yz', 'l', 'r', 'lr', 'lzr', 'lyr',
        'lzry', 'lyrz'. Default='ortho'.

    figure : integer or matplotlib figure, optional
        Matplotlib figure used or its number. If None is given, a
        new figure is created.

    axes : matplotlib axes or 4 tuple of float: (xmin, ymin, width, height), \
optional
        The axes, or the coordinates, in matplotlib figure space,
        of the axes used to display the plot. If None, the complete
        figure is used.

    title : string, optional
        The title displayed on the figure.

    Notes
    -----
    The plotted image should in MNI space for this function to work properly.

    This function is deprecated and will be removed in the 0.9.0 release. Use
    plot_markers instead.

    """
    dep_msg = ("This function is deprecated and will be "
               "removed in the 0.9.0 release. Use plot_markers instead.")
    warnings.warn(dep_msg, DeprecationWarning)

    # input validation
    if cmap is None:
        cmap = plt.cm.viridis_r
    elif isinstance(cmap, str):
        cmap = plt.get_cmap(cmap)
    else:
        cmap = cmap

    node_size = (1 / len(node_coords) * 1e4
                 if node_size == 'auto' else node_size)

    node_coords = np.asarray(node_coords)

    if sparse.issparse(adjacency_matrix):
        adjacency_matrix = adjacency_matrix.toarray()

    adjacency_matrix = np.nan_to_num(adjacency_matrix)

    adjacency_matrix_shape = adjacency_matrix.shape
    if (len(adjacency_matrix_shape) != 2 or  # noqa: W504
            adjacency_matrix_shape[0] != adjacency_matrix_shape[1]):
        raise ValueError(
            "'adjacency_matrix' is supposed to have shape (n, n)."
            ' Its shape was {0}'.format(adjacency_matrix_shape))

    node_coords_shape = node_coords.shape
    if len(node_coords_shape) != 2 or node_coords_shape[1] != 3:
        message = (
            "Invalid shape for 'node_coords'. You passed an "
            "'adjacency_matrix' of shape {0} therefore "
            "'node_coords' should be a array with shape ({0[0]}, 3) "
            "while its shape was {1}").format(adjacency_matrix_shape,
                                              node_coords_shape)

        raise ValueError(message)

    if node_coords_shape[0] != adjacency_matrix_shape[0]:
        raise ValueError(
            "Shape mismatch between 'adjacency_matrix' "
            "and 'node_coords'"
            "'adjacency_matrix' shape is {0}, 'node_coords' shape is {1}"
            .format(adjacency_matrix_shape, node_coords_shape))

    if not np.allclose(adjacency_matrix, adjacency_matrix.T, rtol=1e-3):
        raise ValueError("'adjacency_matrix' should be symmetric")

    # For a masked array, masked values are replaced with zeros
    if hasattr(adjacency_matrix, 'mask'):
        if not (adjacency_matrix.mask == adjacency_matrix.mask.T).all():
            raise ValueError(
                "'adjacency_matrix' was masked with a non symmetric mask")
        adjacency_matrix = adjacency_matrix.filled(0)

    # plotting
    region_strength = np.sum(np.abs(adjacency_matrix), axis=0)
    region_strength /= np.sum(region_strength)

    region_idx_sorted = np.argsort(region_strength)[::-1]
    strength_sorted = region_strength[region_idx_sorted]
    coords_sorted = node_coords[region_idx_sorted]

    display = plot_glass_brain(
        None, display_mode=display_mode, figure=figure, axes=axes, title=title
    )

    for coord, region in zip(coords_sorted, strength_sorted):
        color = list(
            cmap((region - strength_sorted.min()) / strength_sorted.max())
        )
        # reduce alpha for the least strong regions
        color[-1] = (
            (region - strength_sorted.min()) *  # noqa: W504
            (1 / (strength_sorted.max() - strength_sorted.min()))
        )
        # make color to be a 2D array
        color = [color]
        display.add_markers(
            [coord], marker_color=color, marker_size=node_size
        )

    if output_file is not None:
        display.savefig(output_file)
        display.close()
        display = None

    return display


def plot_markers(node_values, node_coords, node_size='auto',
                 node_cmap=plt.cm.viridis_r, node_vmin=None, node_vmax=None,
                 node_threshold=None, alpha=0.7, output_file=None,
                 display_mode="ortho", figure=None, axes=None, title=None,
                 annotate=True, black_bg=False, node_kwargs=None,
                 colorbar=True):
    """Plot network nodes (markers) on top of the brain glass schematics.

    Nodes are color coded according to provided nodal measure. Nodal measure
    usually represents some notion of node importance.

    Parameters
    ----------
    node_values : array_like of length n
        Vector containing nodal importance measure. Each node will be colored
        acording to corresponding node value.

    node_coords : numpy array_like of shape (n, 3)
        3d coordinates of the graph nodes in world space.

    node_size : 'auto' or scalar or array-like, optional
        Size(s) of the nodes in points^2. By default the size of the node is
        inversely proportional to the number of nodes.

    node_cmap : str or colormap, optional
        Colormap used to represent the node measure. Default=plt.cm.viridis_r.

    node_vmin : float, optional
        Lower bound of the colormap. If `None`, the min of the node_values is
        used.

    node_vmax : float, optional
        Upper bound of the colormap. If `None`, the min of the node_values is
        used.

    node_threshold : float
        If provided only the nodes with a value greater than node_threshold
        will be shown.

    alpha : float between 0 and 1, optional
        Alpha transparency for markers. Default=0.7.

    output_file : string, or None, optional
        The name of an image file to export the plot to. Valid extensions
        are .png, .pdf, .svg. If output_file is not None, the plot
        is saved to a file, and the display is closed.

    display_mode : string, optional
        Choose the direction of the cuts: 'x' - sagittal, 'y' - coronal,
        'z' - axial, 'l' - sagittal left hemisphere only,
        'r' - sagittal right hemisphere only, 'ortho' - three cuts are
        performed in orthogonal directions. Possible values are: 'ortho',
        'x', 'y', 'z', 'xz', 'yx', 'yz', 'l', 'r', 'lr', 'lzr', 'lyr',
        'lzry', 'lyrz'. Default='ortho'.

    figure : integer or matplotlib figure, optional
        Matplotlib figure used or its number. If None is given, a
        new figure is created.

    axes : matplotlib axes or 4 tuple of float: (xmin, ymin, width, height), optional
        The axes, or the coordinates, in matplotlib figure space,
        of the axes used to display the plot. If None, the complete
        figure is used.

    title : string, optional
        The title displayed on the figure.

    annotate : boolean, optional
        If annotate is True, positions and left/right annotation
        are added to the plot. Default=True.

    black_bg : boolean, optional
        If True, the background of the image is set to be black. If
        you wish to save figures with a black background, you
        will need to pass "facecolor='k', edgecolor='k'"
        to matplotlib.pyplot.savefig. Default=False.

    node_kwargs : dict, optional
        will be passed as kwargs to the plt.scatter call that plots all
        the nodes in one go

    colorbar : boolean, optional
        If True, display a colorbar on the right of the plots.
        Default=True.

    """
    node_values = np.squeeze(np.array(node_values))
    node_coords = np.array(node_coords)

    # Validate node_values
    if node_values.shape != (node_coords.shape[0], ):
        msg = ("Dimension mismatch: 'node_values' should be vector of length "
               "{0}, but current shape is {1} instead of {2}").format(
                   len(node_coords),
                   node_values.shape,
                   (node_coords.shape[0], ))
        raise ValueError(msg)

    display = plot_glass_brain(None, display_mode=display_mode,
                               figure=figure, axes=axes, title=title,
                               annotate=annotate, black_bg=black_bg)

    if isinstance(node_size, str) and node_size == 'auto':
        node_size = min(1e4 / len(node_coords), 100)

    # Filter out nodes with node values below threshold
    if node_threshold is not None:
        if node_threshold > np.max(node_values):
            msg = ("Provided 'node_threshold' value: {0} should not exceed "
                   "highest node value: {1}").format(node_threshold,
                                                     np.max(node_values))
            raise ValueError(msg)

        retained_nodes = node_values > node_threshold
        node_values = node_values[retained_nodes]
        node_coords = node_coords[retained_nodes]
        if isinstance(node_size, collections.abc.Iterable):
            node_size = [size for ok_retain, size in
                         zip(retained_nodes, node_size) if ok_retain]

    # Calculate node colors based on value
    node_vmin = np.min(node_values) if node_vmin is None else node_vmin
    node_vmax = np.max(node_values) if node_vmax is None else node_vmax
    if node_vmin == node_vmax:
        node_vmin = 0.9 * node_vmin
        node_vmax = 1.1 * node_vmax
    norm = matplotlib.colors.Normalize(vmin=node_vmin, vmax=node_vmax)
    node_cmap = (plt.get_cmap(node_cmap) if isinstance(node_cmap, str)
                 else node_cmap)
    node_color = [node_cmap(norm(node_value)) for node_value in node_values]

    # Prepare additional parameters for plt.scatter
    node_kwargs = {} if node_kwargs is None else node_kwargs
    node_kwargs.update([('alpha', alpha)])

    display.add_markers(
        marker_coords=node_coords,
        marker_color=node_color,
        marker_size=node_size,
        **node_kwargs
    )

    if colorbar:
        display._colorbar = True
        display._show_colorbar(cmap=node_cmap, norm=norm)

    if output_file is not None:
        display.savefig(output_file)
        display.close()
        display = None

    return display


def plot_carpet(img, mask_img=None, mask_labels=None,
                detrend=True, output_file=None,
                figure=None, axes=None, vmin=None, vmax=None, title=None,
                cmap=plt.cm.gist_ncar):
    """Plot an image representation of voxel intensities across time.

    This figure is also known as a "grayplot" or "Power plot".

    Parameters
    ----------
    img : Niimg-like object
        4D image.
        See http://nilearn.github.io/manipulating_images/input_output.html.

    mask_img : Niimg-like object or None, optional
        Limit plotted voxels to those inside the provided mask (default is
        None). If a 3D atlas is provided, voxels will be grouped by atlas
        value and a colorbar will be added to the left side of the figure
        with atlas labels.
        If not specified, a new mask will be derived from data.
        See http://nilearn.github.io/manipulating_images/input_output.html.

    mask_labels : :obj:`dict`, optional
        If ``mask_img`` corresponds to an atlas, then this dictionary maps
        values from the ``mask_img`` to labels. Dictionary keys are labels
        and values are values within the atlas.

    detrend : :obj:`bool`, optional
        Detrend and z-score the data prior to plotting. Default=True.

    output_file : :obj:`str` or None, optional
        The name of an image file to which to export the plot (default is
        None). Valid extensions are .png, .pdf, and .svg.
        If `output_file` is not None, the plot is saved to a file, and the
        display is closed.

    figure : :class:`matplotlib.figure.Figure` or None, optional
        Matplotlib figure used (default is None).
        If None is given, a new figure is created.

    axes : matplotlib axes or None, optional
        The axes used to display the plot (default is None).
        If None, the complete figure is used.

    vmin : float or None, optional
        Lower bound for plotting, passed to matplotlib.pyplot.imshow.
        If None, vmin will be automatically determined based on the data.
        Default=None.

    vmax : float or None, optional
        Upper bound for plotting, passed to matplotlib.pyplot.imshow.
        If None, vmax will be automatically determined based on the data.
        Default=None.

    title : :obj:`str` or None, optional
        The title displayed on the figure (default is None).

    cmap : matplotlib colormap, optional
        The colormap for the sidebar, if an atlas is used.
        Default=plt.cm.gist_ncar.

    Returns
    -------
    figure : :class:`matplotlib.figure.Figure`
        Figure object with carpet plot.

    Notes
    -----
    This figure was originally developed in :footcite:`POWER2017150`.

    In cases of long acquisitions (>800 volumes), the data will be downsampled
    to have fewer than 800 volumes before being plotted.

    References
    ----------
    .. footbibliography::

    """
    img = _utils.check_niimg_4d(img, dtype='auto')

    # Define TR and number of frames
    tr = img.header.get_zooms()[-1]
    n_tsteps = img.shape[-1]

    if mask_img is None:
        mask_img = compute_epi_mask(img)
    else:
        mask_img = _utils.check_niimg_3d(mask_img, dtype='auto')

    is_atlas = len(np.unique(mask_img.get_fdata())) > 2
    if is_atlas:
        background_label = 0

        atlas_img_res = resample_to_img(
            mask_img,
            img,
            interpolation='nearest',
        )
        atlas_bin = math_img(
            'img != {}'.format(background_label),
            img=atlas_img_res,
        )
        masker = NiftiMasker(atlas_bin, target_affine=img.affine)

        data = masker.fit_transform(img)
        atlas_values = masker.transform(atlas_img_res)
        atlas_values = np.squeeze(atlas_values)

        if mask_labels:
            label_dtype = type(list(mask_labels.values())[0])
            if label_dtype != atlas_values.dtype:
                print('Coercing atlas_values to {}'.format(label_dtype))
                atlas_values = atlas_values.astype(label_dtype)

        # Sort data and atlas by atlas values
        order = np.argsort(atlas_values)
        order = np.squeeze(order)
        atlas_values = atlas_values[order]
        data = data[:, order]
    else:
        data = apply_mask(img, mask_img)

    # Detrend and standardize data
    if detrend:
        data = clean(data, t_r=tr, detrend=True, standardize='zscore')

    if figure is None:
        if not axes:
            figsize = (10, 5)
            figure = plt.figure(figsize=figsize)
        else:
            figure = axes.figure

    if axes is None:
        axes = figure.add_subplot(1, 1, 1)
    else:
        assert axes.figure is figure, ('The axes passed are not in the figure')

    # Determine vmin and vmax based on the full data
    std = np.mean(data.std(axis=0))
    default_vmin = data.mean() - (2 * std)
    default_vmax = data.mean() + (2 * std)

    # Avoid segmentation faults for long acquisitions by decimating the data
    LONG_CUTOFF = 800
    # Get smallest power of 2 greater than the number of volumes divided by the
    # cutoff, to determine how much to decimate (downsample) the data.
    n_decimations = int(np.ceil(np.log2(np.ceil(n_tsteps / LONG_CUTOFF))))
    data = data[::2 ** n_decimations, :]

    if is_atlas:
        # Define nested GridSpec
        legend = False
        wratios = [2, 100, 20]
        gs = mgs.GridSpecFromSubplotSpec(
            1,
            2 + int(legend),
            subplot_spec=axes,
            width_ratios=wratios[: 2 + int(legend)],
            wspace=0.0,
        )

        ax0 = plt.subplot(gs[0])
        ax0.set_xticks([])
        ax0.imshow(
            atlas_values[:, np.newaxis],
            interpolation='none',
            aspect='auto',
            cmap=cmap
        )
        if mask_labels:
            # Add labels to middle of each associated band
            mask_labels_inv = {v: k for k, v in mask_labels.items()}
            ytick_locs = [
                np.mean(np.where(atlas_values == i)[0])
                for i in np.unique(atlas_values)
            ]
            ax0.set_yticks(ytick_locs)
            ax0.set_yticklabels([
                mask_labels_inv[i] for i in np.unique(atlas_values)
            ])
        else:
            ax0.set_yticks([])

        # Carpet plot
        axes = plt.subplot(gs[1])  # overwrite axes
        axes.imshow(
            data.T,
            interpolation='nearest',
            aspect='auto',
            cmap='gray',
            vmin=vmin or default_vmin,
            vmax=vmax or default_vmax,
        )
        ax0.tick_params(axis='both', which='both', length=0)
    else:
        axes.imshow(
            data.T,
            interpolation='nearest',
            aspect='auto',
            cmap='gray',
            vmin=vmin or default_vmin,
            vmax=vmax or default_vmax,
        )

    axes.grid(False)
    axes.set_yticks([])
    axes.set_yticklabels([])

    # Set 10 frame markers in X axis
    interval = max(
        (int(data.shape[0] + 1) // 10, int(data.shape[0] + 1) // 5, 1))
    xticks = list(range(0, data.shape[0])[::interval])
    axes.set_xticks(xticks)
    axes.set_xlabel('time (s)')

    if title:
        axes.set_title(title)

    labels = tr * (np.array(xticks))
    labels *= (2 ** n_decimations)
    axes.set_xticklabels(['%.02f' % t for t in labels.tolist()])

    # Remove and redefine spines
    for side in ['top', 'right']:
        # Toggle the spine objects
        axes.spines[side].set_color('none')
        axes.spines[side].set_visible(False)

    axes.xaxis.set_ticks_position('bottom')
    axes.spines['bottom'].set_position(('outward', 10))

    if not mask_labels:
        axes.yaxis.set_ticks_position('left')
        buffer = 20 if is_atlas else 10
        axes.spines['left'].set_position(('outward', buffer))
        axes.set_ylabel('voxels')

    if output_file is not None:
        figure.savefig(output_file)
        plt.close(figure)
        figure = None

    return figure


def plot_img_comparison(ref_imgs, src_imgs, masker, plot_hist=True, log=True,
                        ref_label="image set 1", src_label="image set 2",
                        output_dir=None, axes=None):
    """Creates plots to compare two lists of images and measure correlation.

    The first plot displays linear correlation between voxel values.
    The second plot superimposes histograms to compare values distribution.

    Parameters
    ----------
    ref_imgs : nifti_like
        Reference images.

    src_imgs : nifti_like
        Source images.

    masker : NiftiMasker object
        Mask to be used on data.

    plot_hist : Boolean, optional
        If True then histograms of each img in ref_imgs will be plotted
        along-side the histogram of the corresponding image in src_imgs.
        Default=True.

    log : Boolean, optional
        Passed to plt.hist. Default=True.

    ref_label : str, optional
        Name of reference images. Default='image set 1'.

    src_label : str, optional
        Name of source images. Default='image set 2'.

    output_dir : string, optional
        Directory where plotted figures will be stored.

    axes : list of two matplotlib Axes objects, optional
        Can receive a list of the form [ax1, ax2] to render the plots.
        By default new axes will be created.

    Returns
    -------
    corrs : numpy.ndarray
        Pearson correlation between the images.

    """
    # note: doesn't work with 4d images;
    # when plot_hist is False creates two empty axes and doesn't plot anything
    corrs = []
    for i, (ref_img, src_img) in enumerate(zip(ref_imgs, src_imgs)):
        if axes is None:
            _, (ax1, ax2) = plt.subplots(1, 2, figsize=(12, 5))
        else:
            (ax1, ax2) = axes
        ref_data = masker.transform(ref_img).ravel()
        src_data = masker.transform(src_img).ravel()
        if ref_data.shape != src_data.shape:
            warnings.warn("Images are not shape-compatible")
            return

        corr = stats.pearsonr(ref_data, src_data)[0]
        corrs.append(corr)

        if plot_hist:
            ax1.scatter(
                ref_data, src_data, label="Pearsonr: %.2f" % corr, c="g",
                alpha=.6)
            x = np.linspace(*ax1.get_xlim(), num=100)
            ax1.plot(x, x, linestyle="--", c="k")
            ax1.grid("on")
            ax1.set_xlabel(ref_label)
            ax1.set_ylabel(src_label)
            ax1.legend(loc="best")

            ax2.hist(ref_data, alpha=.6, bins=128, log=log, label=ref_label)
            ax2.hist(src_data, alpha=.6, bins=128, log=log, label=src_label)
            ax2.set_title("Histogram of imgs values")
            ax2.grid("on")
            ax2.legend(loc="best")

            if output_dir is not None:
                if not os.path.exists(output_dir):
                    os.makedirs(output_dir)
                plt.savefig(os.path.join(output_dir, "%04i.png" % i))

        plt.tight_layout()

    return corrs



import warnings
from io import BytesIO
import base64

import numpy as np
from matplotlib import pyplot as plt
import pytest

from nibabel import Nifti1Image

from nilearn import datasets, image
from nilearn.plotting import html_stat_map
from nilearn.image import new_img_like
from nilearn.image import get_data
from ..js_plotting_utils import colorscale


def _check_html(html_view, title=None):
    """ Check the presence of some expected code in the html viewer
    """
    assert isinstance(html_view, html_stat_map.StatMapView)
    assert "var brain =" in str(html_view)
    assert "overlayImg" in str(html_view)
    if title is not None:
        assert "<title>{}</title>".format(title) in str(html_view)


def _simulate_img(affine=np.eye(4)):
    """ Simulate data with one "spot"
        Returns: img, data
    """
    data = np.zeros([8, 8, 8])
    data[4, 4, 4] = 1
    img = Nifti1Image(data, affine)
    return img, data


def _check_affine(affine):
    """ Check positive, isotropic, near-diagonal affine.
    """
    assert(affine[0, 0] == affine[1, 1])
    assert(affine[2, 2] == affine[1, 1])
    assert(affine[0, 0] > 0)

    A, b = image.resampling.to_matrix_vector(affine)
    assert np.all((np.abs(A) > 0.001).sum(axis=0) == 1), (
        "the affine transform was not near-diagonal")


def test_data_to_sprite():

    # Simulate data and turn into sprite
    data = np.zeros([8, 8, 8])
    data[2:6, 2:6, 2:6] = 1
    sprite = html_stat_map._data_to_sprite(data)

    # Generate ground truth for the sprite
    Z = np.zeros([8, 8])
    Zr = np.zeros([2, 8])
    Cr = np.tile(np.array([[0, 0, 1, 1, 1, 1, 0, 0]]), [4, 1])
    C = np.concatenate((Zr, Cr, Zr), axis=0)
    gtruth = np.concatenate((np.concatenate((Z, Z, C), axis=1),
                             np.concatenate((C, C, C), axis=1),
                             np.concatenate((Z, Z, Z), axis=1)),
                            axis=0)

    assert sprite.shape == gtruth.shape, "shape of sprite not as expected"
    assert (sprite == gtruth).all(), "simulated sprite not as expected"


def test_threshold_data():

    data = np.arange(-3, 4)

    # Check that an 'auto' threshold leaves at least one element
    data_t, mask, thresh = html_stat_map._threshold_data(data, threshold='auto')
    gtruth_m = np.array([False, True, True, True, True, True, False])
    gtruth_d = np.array([-3, 0, 0, 0, 0, 0, 3])
    assert (mask == gtruth_m).all()
    assert (data_t == gtruth_d).all()

    # Check that threshold=None keeps everything
    data_t, mask, thresh = html_stat_map._threshold_data(data, threshold=None)
    assert np.all(np.logical_not(mask))
    assert np.all(data_t == data)

    # Check positive threshold works
    data_t, mask, thresh = html_stat_map._threshold_data(data, threshold=1)
    gtruth = np.array([False, False, True, True, True, False, False])
    assert (mask == gtruth).all()

    # Check 0 threshold works
    data_t, mask, thresh = html_stat_map._threshold_data(data, threshold=0)
    gtruth = np.array([False, False, False, True, False, False, False])
    assert (mask == gtruth).all()

    # Check that overly lenient threshold returns array
    data = np.arange(3, 10)
    data_t, mask, thresh = html_stat_map._threshold_data(data, threshold=2)
    gtruth = np.full(7, False)
    assert (mask == gtruth).all()


def test_save_sprite():
    """This test covers _save_sprite as well as _bytesIO_to_base64
    """

    # Generate a simulated volume with a square inside
    data = np.random.RandomState(42).uniform(size=140).reshape(7, 5, 4)
    mask = np.zeros((7, 5, 4), dtype=int)
    mask[1:-1, 1:-1, 1:-1] = 1
    # Save the sprite using BytesIO
    sprite_io = BytesIO()
    html_stat_map._save_sprite(data, sprite_io, vmin=0, vmax=1,
                               mask=mask, format='png')

    # Load the sprite back in base64
    sprite_base64 = html_stat_map._bytesIO_to_base64(sprite_io)

    decoded_io = BytesIO()
    decoded_io.write(base64.b64decode(sprite_base64))
    decoded_io.seek(0)
    img = plt.imread(decoded_io, format="png")
    correct_img = np.ma.array(html_stat_map._data_to_sprite(data),
                              mask=html_stat_map._data_to_sprite(mask))
    correct_img = plt.Normalize(0, 1)(correct_img)
    cmapped = plt.get_cmap("Greys")(correct_img)
    assert np.allclose(img, cmapped, atol=.1)


@pytest.mark.parametrize("cmap", ["tab10", "cold_hot"])
@pytest.mark.parametrize("n_colors", [7, 20])
def test_save_cmap(cmap, n_colors):
    """This test covers _save_cmap as well as _bytesIO_to_base64
    """
    # Save the cmap using BytesIO
    cmap_io = BytesIO()
    html_stat_map._save_cm(cmap_io, cmap, format='png', n_colors=n_colors)

    # Load the colormap back in base64
    cmap_base64 = html_stat_map._bytesIO_to_base64(cmap_io)

    decoded_io = BytesIO()
    decoded_io.write(base64.b64decode(cmap_base64))
    decoded_io.seek(0)
    img = plt.imread(decoded_io, format="png")
    expected = plt.get_cmap(cmap)(np.linspace(0, 1, n_colors))
    assert np.allclose(img, expected, atol=.1)


def test_mask_stat_map():

    # Generate simple simulated data with one "spot"
    img, data = _simulate_img()

    # Try not to threshold anything
    mask_img, img, data_t, thre = html_stat_map._mask_stat_map(img,
                                                               threshold=None)
    assert np.max(get_data(mask_img)) == 0

    # Now threshold at zero
    mask_img, img, data_t, thre = html_stat_map._mask_stat_map(img,
                                                               threshold=0)
    assert np.min((data == 0) == get_data(mask_img))


def test_load_bg_img():

    # Generate simple simulated data with non-diagonal affine
    affine = np.eye(4)
    affine[0, 0] = -1
    affine[0, 1] = 0.1
    img, data = _simulate_img(affine)

    # use empty bg_img
    bg_img, bg_min, bg_max, black_bg = html_stat_map._load_bg_img(img,
                                                                  bg_img=None)
    # Check positive isotropic, near-diagonal affine
    _check_affine(bg_img.affine)

    # Try to load the default background
    bg_img, bg_min, bg_max, black_bg = html_stat_map._load_bg_img(img)

    # Check positive isotropic, near-diagonal affine
    _check_affine(bg_img.affine)


def test_resample_stat_map():

    # Start with simple simulated data
    bg_img, data = _simulate_img()

    # Now double the voxel size and mess with the affine
    affine = 2 * np.eye(4)
    affine[3, 3] = 1
    affine[0, 1] = 0.1
    stat_map_img = Nifti1Image(data, affine)

    # Make a mask for the stat image
    mask_img = new_img_like(stat_map_img, data > 0, stat_map_img.affine)

    # Now run the resampling
    stat_map_img, mask_img = html_stat_map._resample_stat_map(
        stat_map_img, bg_img, mask_img, resampling_interpolation='nearest')

    # Check positive isotropic, near-diagonal affine
    _check_affine(stat_map_img.affine)
    _check_affine(mask_img.affine)

    # Check voxel size matches bg_img
    assert stat_map_img.affine[0, 0] == bg_img.affine[0, 0], (
        "stat_map_img was not resampled at the resolution of background")
    assert mask_img.affine[0, 0] == bg_img.affine[0, 0], (
        "mask_img was not resampled at the resolution of background")


def test_json_view_params():

    # Try to generate some sprite parameters
    params = html_stat_map._json_view_params(
        shape=[4, 4, 4], affine=np.eye(4), vmin=0, vmax=1,
        cut_slices=[1, 1, 1], black_bg=True, opacity=0.5, draw_cross=False,
        annotate=True, title="A test", colorbar=True, value=True)

    # Just check that a structure was generated,
    # and test a single parameter
    assert params['overlay']['opacity'] == 0.5


def test_json_view_size():

    # Build some minimal sprite Parameters
    sprite_params = {'nbSlice': {'X': 4, 'Y': 4, 'Z': 4}}
    width, height = html_stat_map._json_view_size(sprite_params)

    # This is a simple case: height is 4 pixels, width 3 x 4 = 12 pixels
    # with an additional 120% height factor for annotations and margins
    ratio = 1.2 * 4 / 12

    # check we received the expected width and height
    width_exp = 600
    height_exp = np.ceil(ratio * 600)
    assert width == width_exp, "html viewer does not have expected width"
    assert height == height_exp, "html viewer does not have expected height"


def test_json_view_data():

    # simple simulated data for stat_img and background
    bg_img, data = _simulate_img()
    stat_map_img, data = _simulate_img()

    # make a mask
    mask_img = new_img_like(stat_map_img, data > 0, stat_map_img.affine)

    # Get color bar and data ranges
    colors = colorscale('cold_hot', data.ravel(), threshold=0,
                        symmetric_cmap=True, vmax=1)

    # Build a sprite
    json_view = html_stat_map._json_view_data(
        bg_img, stat_map_img, mask_img, bg_min=0, bg_max=1, colors=colors,
        cmap='cold_hot', colorbar=True)

    # Check the presence of critical fields
    assert isinstance(json_view['bg_base64'], str)
    assert isinstance(json_view['stat_map_base64'], str)
    assert isinstance(json_view['cm_base64'], str)

    return json_view, data


def test_json_view_to_html():

    # Re use the data simulated in another test
    json_view, data = test_json_view_data()
    json_view['params'] = html_stat_map._json_view_params(
        data.shape, np.eye(4), vmin=0, vmax=1, cut_slices=[1, 1, 1],
        black_bg=True, opacity=1, draw_cross=True, annotate=False,
        title="test", colorbar=True)

    # Create a viewer
    html_view = html_stat_map._json_view_to_html(json_view)
    _check_html(html_view)


def test_get_cut_slices():

    # Generate simple simulated data with one "spot"
    img, data = _simulate_img()

    # Use automatic selection of coordinates
    cut_slices = html_stat_map._get_cut_slices(img, cut_coords=None,
                                               threshold=None)
    assert (cut_slices == [4, 4, 4]).all()

    # Check that using a single number for cut_coords raises an error
    with pytest.raises(ValueError):
        html_stat_map._get_cut_slices(img, cut_coords=4, threshold=None)

    # Check that it is possible to manually specify coordinates
    cut_slices = html_stat_map._get_cut_slices(img, cut_coords=[2, 2, 2],
                                               threshold=None)
    assert (cut_slices == [2, 2, 2]).all()

    # Check that the affine does not change where the cut is done
    affine = 2 * np.eye(4)
    img = Nifti1Image(data, affine)
    cut_slices = html_stat_map._get_cut_slices(img, cut_coords=None,
                                               threshold=None)
    assert (cut_slices == [4, 4, 4]).all()


def test_view_img():
    mni = datasets.load_mni152_template()
    with warnings.catch_warnings(record=True) as w:
        # Create a fake functional image by resample the template
        img = image.resample_img(mni, target_affine=3 * np.eye(3))
        html_view = html_stat_map.view_img(img)
        _check_html(html_view, title="Slice viewer")
        html_view = html_stat_map.view_img(img, threshold='95%',
                                           title="SOME_TITLE")
        _check_html(html_view, title="SOME_TITLE")
        html_view = html_stat_map.view_img(img, bg_img=mni)
        _check_html(html_view)
        html_view = html_stat_map.view_img(img, bg_img=None)
        _check_html(html_view)
        html_view = html_stat_map.view_img(img, threshold=2., vmax=4.)
        _check_html(html_view)
        html_view = html_stat_map.view_img(img, symmetric_cmap=False)
        img_4d = image.new_img_like(img, get_data(img)[:, :, :, np.newaxis])
        assert len(img_4d.shape) == 4
        html_view = html_stat_map.view_img(img_4d, threshold=2., vmax=4.)
        _check_html(html_view)
        html_view = html_stat_map.view_img(img_4d, threshold=1e6)
        _check_html(html_view)

    # Check that all warnings were expected
    warnings_set = set(warning_.category for warning_ in w)
    expected_set = set([FutureWarning, UserWarning,
                       DeprecationWarning])
    assert warnings_set.issubset(expected_set), (
        "the following warnings were not expected: {}").format(
        warnings_set.difference(expected_set))



"""
This module is for contrast computation and operation on contrast to
obtain fixed effect results.

Author: Bertrand Thirion, Martin Perez-Guevara, Ana Luisa Pinho 2020
"""

from warnings import warn

import numpy as np
import scipy.stats as sps
import pandas as pd

from nilearn.input_data import NiftiMasker
from nilearn._utils.glm import z_score

DEF_TINY = 1e-50
DEF_DOFMAX = 1e10


def expression_to_contrast_vector(expression, design_columns):
    """ Converts a string describing a contrast to a contrast vector

    Parameters
    ----------
    expression : string
        The expression to convert to a vector.

    design_columns : list or array of strings
        The column names of the design matrix.

    Notes
    -----
    This function is experimental.
    It may change in any future release of Nilearn.

    """
    if expression in design_columns:
        contrast_vector = np.zeros(len(design_columns))
        contrast_vector[list(design_columns).index(expression)] = 1.
        return contrast_vector
    df = pd.DataFrame(np.eye(len(design_columns)), columns=design_columns)
    contrast_vector = df.eval(expression, engine="python").values
    return contrast_vector


def compute_contrast(labels, regression_result, con_val, contrast_type=None):
    """ Compute the specified contrast given an estimated glm

    Parameters
    ----------
    labels : array of shape (n_voxels,)
        A map of values on voxels used to identify the corresponding model

    regression_result : dict
        With keys corresponding to the different labels
        values are RegressionResults instances corresponding to the voxels.

    con_val : numpy.ndarray of shape (p) or (q, p)
        Where q = number of contrast vectors and p = number of regressors.

    contrast_type : {None, 't', 'F'}, optional
        Type of the contrast.  If None, then defaults to 't' for 1D
        `con_val` and 'F' for 2D `con_val`

    Returns
    -------
    con : Contrast instance,
        Yields the statistics of the contrast (effects, variance, p-values)

    Notes
    -----
    This function is experimental.
    It may change in any future release of Nilearn.

    """
    con_val = np.asarray(con_val)
    dim = 1
    if con_val.ndim > 1:
        dim = con_val.shape[0]

    if contrast_type is None:
        contrast_type = 't' if dim == 1 else 'F'

    acceptable_contrast_types = ['t', 'F']
    if contrast_type not in acceptable_contrast_types:
        raise ValueError(
            '"{0}" is not a known contrast type. Allowed types are {1}'.
                format(contrast_type, acceptable_contrast_types))

    if contrast_type == 't':
        effect_ = np.zeros((1, labels.size))
        var_ = np.zeros(labels.size)
        for label_ in regression_result:
            label_mask = labels == label_
            resl = regression_result[label_].Tcontrast(con_val)
            effect_[:, label_mask] = resl.effect.T
            var_[label_mask] = (resl.sd ** 2).T
    elif contrast_type == 'F':
        from scipy.linalg import sqrtm
        effect_ = np.zeros((dim, labels.size))
        var_ = np.zeros(labels.size)
        for label_ in regression_result:
            label_mask = labels == label_
            reg = regression_result[label_]
            cbeta = np.atleast_2d(np.dot(con_val, reg.theta))
            invcov = np.linalg.inv(np.atleast_2d(
                reg.vcov(matrix=con_val, dispersion=1.0)))
            wcbeta = np.dot(sqrtm(invcov), cbeta)
            rss = reg.dispersion
            effect_[:, label_mask] = wcbeta
            var_[label_mask] = rss

    dof_ = regression_result[label_].df_residuals
    return Contrast(effect=effect_, variance=var_, dim=dim, dof=dof_,
                    contrast_type=contrast_type)


def _compute_fixed_effect_contrast(labels, results, con_vals,
                                   contrast_type=None):
    """Computes the summary contrast assuming fixed effects.

    Adds the same contrast applied to all labels and results lists.

    """
    contrast = None
    n_contrasts = 0
    for i, (lab, res, con_val) in enumerate(zip(labels, results, con_vals)):
        if np.all(con_val == 0):
            warn('Contrast for session %d is null' % i)
            continue
        contrast_ = compute_contrast(lab, res, con_val, contrast_type)
        if contrast is None:
            contrast = contrast_
        else:
            contrast = contrast + contrast_
        n_contrasts += 1
    if contrast is None:
        raise ValueError('all contrasts provided were null contrasts')
    return contrast * (1. / n_contrasts)


class Contrast(object):
    """ The contrast class handles the estimation of statistical contrasts
    on a given model: student (t) or Fisher (F).
    The important feature is that it supports addition,
    thus opening the possibility of fixed-effects models.

    The current implementation is meant to be simple,
    and could be enhanced in the future on the computational side
    (high-dimensional F constrasts may lead to memory breakage).

    """
    def __init__(self, effect, variance, dim=None, dof=DEF_DOFMAX,
                 contrast_type='t', tiny=DEF_TINY, dofmax=DEF_DOFMAX):
        """
        Parameters
        ----------
        effect : array of shape (contrast_dim, n_voxels)
            The effects related to the contrast.

        variance : array of shape (n_voxels)
            The associated variance estimate.

        dim : int or None, optional
            The dimension of the contrast.

        dof : scalar, optional
            The degrees of freedom of the residuals.
            Default=DEF_DOFMAX

        contrast_type : {'t', 'F'}, optional
            Specification of the contrast type.
            Default='t'.

        tiny : float, optional
            Small quantity used to avoid numerical underflows.
            Default=DEF_TINY

        dofmax : scalar, optional
            The maximum degrees of freedom of the residuals.
            Default=DEF_DOFMAX.

        Warnings
        --------
        This class is experimental.
        It may change in any future release of Nilearn.

        """
        if variance.ndim != 1:
            raise ValueError('Variance array should have 1 dimension')
        if effect.ndim != 2:
            raise ValueError('Effect array should have 2 dimensions')

        self.effect = effect
        self.variance = variance
        self.dof = float(dof)
        if dim is None:
            self.dim = effect.shape[0]
        else:
            self.dim = dim
        if self.dim > 1 and contrast_type == 't':
            print('Automatically converted multi-dimensional t to F contrast')
            contrast_type = 'F'
        self.contrast_type = contrast_type
        self.stat_ = None
        self.p_value_ = None
        self.one_minus_pvalue_ = None
        self.baseline = 0
        self.tiny = tiny
        self.dofmax = dofmax

    def effect_size(self):
        """Make access to summary statistics more straightforward when
        computing contrasts"""
        return self.effect[0, :]

    def effect_variance(self):
        """Make access to summary statistics more straightforward when
        computing contrasts"""
        return self.variance

    def stat(self, baseline=0.0):
        """Return the decision statistic associated with the test of the
        null hypothesis: (H0) 'contrast equals baseline'

        Parameters
        ----------
        baseline : float, optional
            Baseline value for the test statistic.
            Default=0.0.

        Returns
        -------
        stat : 1-d array, shape=(n_voxels,)
            statistical values, one per voxel.

        """
        self.baseline = baseline

        # Case: one-dimensional contrast ==> t or t**2
        if self.contrast_type == 'F':
            stat = np.sum((self.effect - baseline) ** 2, 0) / self.dim / \
                   np.maximum(self.variance, self.tiny)
        elif self.contrast_type == 't':
            # avoids division by zero
            stat = (self.effect - baseline) / np.sqrt(
                np.maximum(self.variance, self.tiny))
        else:
            raise ValueError('Unknown statistic type')
        self.stat_ = stat.ravel()
        return self.stat_

    def p_value(self, baseline=0.0):
        """Return a parametric estimate of the p-value associated with
        the null hypothesis (H0): 'contrast equals baseline',
        using the survival function

        Parameters
        ----------
        baseline : float, optional
            Baseline value for the test statistic.
            Default=0.0.

        Returns
        -------
        p_values : 1-d array, shape=(n_voxels,)
            p-values, one per voxel

        """
        if self.stat_ is None or not self.baseline == baseline:
            self.stat_ = self.stat(baseline)
        # Valid conjunction as in Nichols et al, Neuroimage 25, 2005.
        if self.contrast_type == 't':
            p_values = sps.t.sf(self.stat_, np.minimum(self.dof, self.dofmax))
        elif self.contrast_type == 'F':
            p_values = sps.f.sf(self.stat_, self.dim, np.minimum(
                self.dof, self.dofmax))
        else:
            raise ValueError('Unknown statistic type')
        self.p_value_ = p_values
        return p_values

    def one_minus_pvalue(self, baseline=0.0):
        """Return a parametric estimate of the 1 - p-value associated with
        the null hypothesis (H0): 'contrast equals baseline',
        using the cumulative distribution function,
        to ensure numerical stability

        Parameters
        ----------
        baseline : float, optional
            Baseline value for the test statistic.
            Default=0.0.

        Returns
        -------
        one_minus_pvalues : 1-d array, shape=(n_voxels,)
            one_minus_pvalues, one per voxel

        """
        if self.stat_ is None or not self.baseline == baseline:
            self.stat_ = self.stat(baseline)
        # Valid conjunction as in Nichols et al, Neuroimage 25, 2005.
        if self.contrast_type == 't':
            one_minus_pvalues = sps.t.cdf(self.stat_,
                                          np.minimum(self.dof, self.dofmax))
        else:
            assert self.contrast_type == 'F'
            one_minus_pvalues = sps.f.cdf(self.stat_, self.dim,
                                          np.minimum(self.dof, self.dofmax))
        self.one_minus_pvalue_ = one_minus_pvalues
        return one_minus_pvalues

    def z_score(self, baseline=0.0):
        """Return a parametric estimation of the z-score associated
        with the null hypothesis: (H0) 'contrast equals baseline'

        Parameters
        ----------
        baseline : float, optional,
            Baseline value for the test statistic.
            Default=0.0.

        Returns
        -------
        z_score : 1-d array, shape=(n_voxels,)
            statistical values, one per voxel

        """
        if self.p_value_ is None or not self.baseline == baseline:
            self.p_value_ = self.p_value(baseline)
        if self.one_minus_pvalue_ is None:
            self.one_minus_pvalue_ = self.one_minus_pvalue(baseline)

        # Avoid inf values kindly supplied by scipy.
        self.z_score_ = z_score(self.p_value_,
                                one_minus_pvalue=self.one_minus_pvalue_)
        return self.z_score_

    def __add__(self, other):
        """Addition of selfwith others, Yields an new Contrast instance
        This should be used only on indepndent contrasts"""
        if self.contrast_type != other.contrast_type:
            raise ValueError(
                'The two contrasts do not have consistant type dimensions')
        if self.dim != other.dim:
            raise ValueError(
                'The two contrasts do not have compatible dimensions')
        dof_ = self.dof + other.dof
        if self.contrast_type == 'F':
            warn('Running approximate fixed effects on F statistics.')
        effect_ = self.effect + other.effect
        variance_ = self.variance + other.variance
        return Contrast(effect=effect_, variance=variance_, dim=self.dim,
                        dof=dof_, contrast_type=self.contrast_type)

    def __rmul__(self, scalar):
        """Multiplication of the contrast by a scalar"""
        scalar = float(scalar)
        effect_ = self.effect * scalar
        variance_ = self.variance * scalar ** 2
        dof_ = self.dof
        return Contrast(effect=effect_, variance=variance_, dof=dof_,
                        contrast_type=self.contrast_type)

    __mul__ = __rmul__

    def __div__(self, scalar):
        return self.__rmul__(1 / float(scalar))


def compute_fixed_effects(contrast_imgs, variance_imgs, mask=None,
                          precision_weighted=False):
    """Compute the fixed effects, given images of effects and variance

    Parameters
    ----------
    contrast_imgs : list of Nifti1Images or strings
        The input contrast images.

    variance_imgs : list of Nifti1Images or strings
        The input variance images.

    mask : Nifti1Image or NiftiMasker instance or None, optional
        Mask image. If None, it is recomputed from contrast_imgs.

    precision_weighted : Bool, optional
        Whether fixed effects estimates should be weighted by inverse
        variance or not. Default=False.

    Returns
    -------
    fixed_fx_contrast_img : Nifti1Image
        The fixed effects contrast computed within the mask.

    fixed_fx_variance_img : Nifti1Image
        The fixed effects variance computed within the mask.

    fixed_fx_t_img : Nifti1Image
        The fixed effects t-test computed within the mask.

    Notes
    -----
    This function is experimental.
    It may change in any future release of Nilearn.

    """
    if len(contrast_imgs) != len(variance_imgs):
        raise ValueError(
            'The number of contrast images (%d) '
            'differs from the number of variance images (%d). '
            % (len(contrast_imgs), len(variance_imgs))
        )

    if isinstance(mask, NiftiMasker):
        masker = mask.fit()
    elif mask is None:
        masker = NiftiMasker().fit(contrast_imgs)
    else:
        masker = NiftiMasker(mask_img=mask).fit()

    variances = masker.transform(variance_imgs)
    contrasts = masker.transform(contrast_imgs)

    (fixed_fx_contrast,
     fixed_fx_variance, fixed_fx_t) = _compute_fixed_effects_params(
        contrasts, variances, precision_weighted)

    fixed_fx_contrast_img = masker.inverse_transform(fixed_fx_contrast)
    fixed_fx_variance_img = masker.inverse_transform(fixed_fx_variance)
    fixed_fx_t_img = masker.inverse_transform(fixed_fx_t)
    return fixed_fx_contrast_img, fixed_fx_variance_img, fixed_fx_t_img


def _compute_fixed_effects_params(contrasts, variances, precision_weighted):
    """ Computes the fixed effects t-statistic, contrast, variance,
    given arrays of effects and variance.
    """
    tiny = 1.e-16
    contrasts, variances = np.asarray(contrasts), np.asarray(variances)
    variances = np.maximum(variances, tiny)

    if precision_weighted:
        weights = 1. / variances
        fixed_fx_variance = 1. / np.sum(weights, 0)
        fixed_fx_contrasts = np.sum(contrasts * weights, 0) * fixed_fx_variance
    else:
        fixed_fx_variance = np.mean(variances, 0) / len(variances)
        fixed_fx_contrasts = np.mean(contrasts, 0)

    fixed_fx_stat = fixed_fx_contrasts / np.sqrt(fixed_fx_variance)
    return fixed_fx_contrasts, fixed_fx_variance, fixed_fx_stat



"""
Brain schematics plotting for glass brain functionality
"""

import json
import os

from distutils.version import LooseVersion

import matplotlib
from matplotlib.path import Path
from matplotlib import patches
from matplotlib import colors
from matplotlib import transforms


def _codes_bezier(pts):
    bezier_num = len(pts)
    # Next two lines are meant to handle both Bezier 3 and 4
    path_attr = 'CURVE{0}'.format(bezier_num)
    codes = [getattr(Path, path_attr)] * (bezier_num - 1)
    return [Path.MOVETO] + codes


def _codes_segment(pts):
    return [Path.MOVETO, Path.LINETO]


def _codes(atype, pts):
    dispatch = {'bezier': _codes_bezier,
                'segment': _codes_segment}

    return dispatch[atype](pts)


def _invert_color(color):
    """Return inverted color

    If color is (R, G, B) it returns (1 - R, 1 - G, 1 - B). If
    'color' can not be converted to a color it is returned
    unmodified.

    """
    try:
        color_converter = colors.ColorConverter()
        color_rgb = color_converter.to_rgb(color)
        return tuple(1 - level for level in color_rgb)
    except ValueError:
        return color


def _get_mpl_patches(json_content, transform=None,
                     invert_color=False, **kwargs):
    """Walks over the json content and builds a list of matplotlib patches
    """
    mpl_patches = []
    kwargs_edgecolor = kwargs.pop('edgecolor', None)
    kwargs_linewidth = kwargs.pop('linewidth', None)
    for path in json_content['paths']:
        if kwargs_edgecolor is not None:
            edgecolor = kwargs_edgecolor
        else:
            edgecolor = path['edgecolor']
            if invert_color:
                edgecolor = _invert_color(edgecolor)
        linewidth = kwargs_linewidth or path['linewidth']
        path_id = path['id']

        for item in path['items']:
            type = item['type']
            pts = item['pts']
            codes = _codes(type, pts)
            path = Path(pts, codes)
            patch = patches.PathPatch(path,
                                      edgecolor=edgecolor,
                                      linewidth=linewidth,
                                      facecolor='none',
                                      gid=path_id,
                                      transform=transform,
                                      **kwargs)

            mpl_patches.append(patch)

    return mpl_patches


def _get_json_and_transform(direction):
    """Returns the json filename and and an affine transform, which has
    been tweaked by hand to fit the MNI template
    """
    direction_to_view_name = {'x': 'side',
                              'y': 'back',
                              'z': 'top',
                              'l': 'side',
                              'r': 'side'}

    direction_to_transform_params = {
        'x': [0.38, 0, 0, 0.38, -108, -70],
        'y': [0.39, 0, 0, 0.39, -73, -73],
        'z': [0.36, 0, 0, 0.37, -71, -107],
        'l': [0.38, 0, 0, 0.38, -108, -70],
        'r': [0.38, 0, 0, 0.38, -108, -70]}

    dirname = os.path.dirname(os.path.abspath(__file__))
    dirname = os.path.join(dirname, 'glass_brain_files')
    direction_to_filename = dict([
        (_direction, os.path.join(
            dirname,
            'brain_schematics_{0}.json'.format(view_name)))
        for _direction, view_name in direction_to_view_name.items()])

    direction_to_transforms = dict([
        (_direction, transforms.Affine2D.from_values(*params))
        for _direction, params in direction_to_transform_params.items()])

    direction_to_json_and_transform = dict([
        (_direction, (direction_to_filename[_direction],
                      direction_to_transforms[_direction]))
        for _direction in direction_to_filename])

    filename_and_transform = direction_to_json_and_transform.get(direction)

    if filename_and_transform is None:
        message = ("No glass brain view associated with direction '{0}'. "
                   "Possible directions are {1}").format(
                       direction,
                       list(direction_to_json_and_transform.keys()))
        raise ValueError(message)

    return filename_and_transform


def _get_object_bounds(json_content, transform):
    xmin, xmax, ymin, ymax = json_content['metadata']['bounds']
    x0, y0 = transform.transform((xmin, ymin))
    x1, y1 = transform.transform((xmax, ymax))

    xmin, xmax = min(x0, x1), max(x0, x1)
    ymin, ymax = min(y0, y1), max(y0, y1)

    # A combination of a proportional factor (fraction of the drawing)
    # and a guestimate of the linewidth
    xmargin = (xmax - xmin) * 0.025 + .1
    ymargin = (ymax - ymin) * 0.025 + .1
    return xmin - xmargin, xmax + xmargin, ymin - ymargin, ymax + ymargin


def plot_brain_schematics(ax, direction, **kwargs):
    """Creates matplotlib patches from a json custom format and plots them
    on a matplotlib Axes.

    Parameters
    ----------
    ax : A MPL axes instance
        The axes in which the plots will be drawn.

    direction : {'x', 'y', 'z', 'l', 'r'}
        The directions of the view.

    **kwargs :
        Passed to the matplotlib patches constructor.

    Returns
    -------
    object_bounds : (xmin, xmax, ymin, ymax) tuple
        Useful for the caller to be able to set axes limits.

    """
    if LooseVersion(matplotlib.__version__) >= LooseVersion("2.0"):
        get_axis_bg_color = ax.get_facecolor()
    else:
        get_axis_bg_color = ax.get_axis_bgcolor()

    black_bg = colors.colorConverter.to_rgba(get_axis_bg_color) \
                    == colors.colorConverter.to_rgba('k')

    json_filename, transform = _get_json_and_transform(direction)
    with open(json_filename) as json_file:
        json_content = json.loads(json_file.read())

    mpl_patches = _get_mpl_patches(json_content,
                                   transform=transform + ax.transData,
                                   invert_color=black_bg,
                                   **kwargs)

    for mpl_patch in mpl_patches:
        ax.add_patch(mpl_patch)

    object_bounds = _get_object_bounds(json_content, transform)

    return object_bounds



from pathlib import Path
import zipfile
import tarfile
import re

import requests

import pytest

from nilearn._utils.data_gen import generate_fake_fmri
from nilearn import image
from nilearn.datasets import _testing


def test_sender_key_order(request_mocker, tmp_path):
    request_mocker.url_mapping["*message.txt"] = "message"
    resp = requests.get("https://example.org/message.txt")
    assert resp.text == "message"
    request_mocker.url_mapping["*.txt"] = "new message"
    resp = requests.get("https://example.org/message.txt")
    assert resp.text == "new message"
    request_mocker.url_mapping["*.csv"] = "other message"
    resp = requests.get("https://example.org/message.txt")
    assert resp.text == "new message"


def test_loading_from_archive_contents(tmp_path):
    expected_contents = sorted(
        [
            Path("README.txt"),
            Path("data"),
            Path("data", "img.nii.gz"),
            Path("data", "labels.csv"),
        ]
    )
    resp = requests.get("https://example.org/example_zip")
    file_path = tmp_path / "archive.zip"
    file_path.write_bytes(resp.content)
    zip_extract_dir = tmp_path / "extract_zip"
    zip_extract_dir.mkdir()
    with zipfile.ZipFile(str(file_path)) as zipf:
        assert sorted(map(Path, zipf.namelist())) == expected_contents
        zipf.extractall(str(zip_extract_dir))
    labels_file = zip_extract_dir / "data" / "labels.csv"
    assert labels_file.read_bytes() == b""
    for url_end in ["_default_format", "_tar_gz"]:
        resp = requests.get("https://example.org/example{}".format(url_end))
        file_path = tmp_path / "archive.tar.gz"
        file_path.write_bytes(resp.content)
        tar_extract_dir = tmp_path / "extract_tar{}".format(url_end)
        tar_extract_dir.mkdir()
        with tarfile.open(str(file_path)) as tarf:
            assert (
                sorted(map(Path, tarf.getnames()))
                == [Path(".")] + expected_contents
            )
            tarf.extractall(str(tar_extract_dir))
        labels_file = tar_extract_dir / "data" / "labels.csv"
        assert labels_file.read_bytes() == b""


def test_sender_regex(request_mocker):
    url = "https://example.org/info?key=value&name=nilearn"
    pattern = re.compile(
        r".*example.org/(?P<section>.*)\?.*name=(?P<name>[^&]+)"
    )
    request_mocker.url_mapping[pattern] = r"in \g<section>: hello \2"
    resp = requests.get(url)
    assert resp.text == "in info: hello nilearn"

    def f(match, request):
        return "name: {}, url: {}".format(match.group("name"), request.url)

    request_mocker.url_mapping[pattern] = f
    resp = requests.get(url)
    assert resp.text == "name: nilearn, url: {}".format(url)

    def g(match, request):
        return 403

    request_mocker.url_mapping[pattern] = g
    resp = requests.get(url)
    with pytest.raises(requests.HTTPError, match="Error"):
        resp.raise_for_status()


def test_sender_status(request_mocker):
    request_mocker.url_mapping["*good"] = 200
    request_mocker.url_mapping["*forbidden"] = 403
    resp = requests.get("https://example.org/good")
    assert resp.status_code == 200
    assert resp.text == "OK"
    resp.raise_for_status()
    resp = requests.get("https://example.org/forbidden")
    assert resp.status_code == 403
    assert resp.text == "ERROR"
    with pytest.raises(requests.HTTPError, match="Error"):
        resp.raise_for_status()


class _MyError(Exception):
    pass


def test_sender_exception(request_mocker):
    request_mocker.url_mapping["*bad"] = _MyError("abc")
    with pytest.raises(_MyError, match="abc"):
        requests.get("ftp:example.org/bad")


def test_sender_img(request_mocker, tmp_path):
    request_mocker.url_mapping["*"] = generate_fake_fmri()[0]
    resp = requests.get("ftp:example.org/download")
    file_path = tmp_path / "img.nii.gz"
    file_path.write_bytes(resp.content)
    img = image.load_img(str(file_path))
    assert img.shape == (10, 11, 12, 17)


class _MyResponse(_testing.Response):
    def json(self):
        return '{"count": 1}'


def test_sender_response(request_mocker):
    request_mocker.url_mapping["*example.org/a"] = _MyResponse("", "")

    def f(match, request):
        resp = _testing.Response(b"hello", request.url)
        resp.headers["cookie"] = "abc"
        return resp

    request_mocker.url_mapping["*example.org/b"] = f
    resp = requests.get("https://example.org/a")
    assert resp.json() == '{"count": 1}'
    resp = requests.get("https://example.org/b")
    assert resp.headers["cookie"] == "abc"


def test_sender_path(request_mocker, tmp_path):
    file_path = tmp_path / "readme.txt"
    with file_path.open("w") as f:
        f.write("hello")
    request_mocker.url_mapping["*path"] = str(file_path)
    request_mocker.url_mapping["*content"] = file_path
    resp = requests.get("https://example.org/path")
    assert resp.text == str(file_path)
    resp = requests.get("https://example.org/content")
    assert resp.text == "hello"


def test_sender_bad_input(request_mocker):
    request_mocker.url_mapping["*"] = 2.5
    with pytest.raises(TypeError):
        requests.get("https://example.org")


def test_dict_to_archive(tmp_path):
    subdir = tmp_path / "tmp"
    subdir.mkdir()
    (subdir / "labels.csv").touch()
    (subdir / "img.nii.gz").touch()
    archive_spec = {
        "empty_data": subdir,
        "empty_data_path.txt": str(subdir),
        Path("data", "labels.csv"): "a,b,c",
        Path("data", "img.nii.gz"): generate_fake_fmri()[0],
        Path("a", "b", "c"): (100).to_bytes(
            length=1, byteorder="big", signed=False
        ),
    }
    targz = _testing.dict_to_archive(archive_spec)
    extract_dir = tmp_path / "extract"
    extract_dir.mkdir()
    archive_path = tmp_path / "archive"
    with archive_path.open("wb") as f:
        f.write(targz)
    with tarfile.open(str(archive_path)) as tarf:
        tarf.extractall(str(extract_dir))
    img = image.load_img(str(extract_dir / "data" / "img.nii.gz"))
    assert img.shape == (10, 11, 12, 17)
    with (extract_dir / "a" / "b" / "c").open("rb") as f:
        assert int.from_bytes(f.read(), byteorder="big", signed=False) == 100
    with open(str(extract_dir / "empty_data" / "labels.csv")) as f:
        assert f.read() == ""
    zip_archive = _testing.dict_to_archive(
        {"readme.txt": "hello", "archive": targz}, "zip"
    )
    with archive_path.open("wb") as f:
        f.write(zip_archive)
    with zipfile.ZipFile(str(archive_path)) as zipf:
        with zipf.open("archive", "r") as f:
            assert f.read() == targz
    from_list = _testing.list_to_archive(archive_spec.keys())
    with archive_path.open("wb") as f:
        f.write(from_list)
    with tarfile.open(str(archive_path)) as tarf:
        assert sorted(map(Path, tarf.getnames())) == sorted(
            list(map(Path, archive_spec.keys()))
            + [Path("."), Path("a"), Path("a", "b"), Path("data")]
        )



""" Test Region Extractor and its functions """

import numpy as np
import nibabel
import pytest
from scipy import ndimage

from nilearn.regions import (connected_regions, RegionExtractor,
                             connected_label_regions)
from nilearn.regions.region_extractor import (_threshold_maps_ratio,
                                              _remove_small_regions)

from nilearn._utils.data_gen import generate_maps, generate_labeled_regions
from nilearn._utils.exceptions import DimensionError
from nilearn.image import get_data


def _make_random_data(shape):
    affine = np.eye(4)
    rng = np.random.RandomState(42)
    data_rng = rng.normal(size=shape)
    img = nibabel.Nifti1Image(data_rng, affine)
    data = get_data(img)
    return img, data


def test_invalid_thresholds_in_threshold_maps_ratio():
    maps, _ = generate_maps((10, 11, 12), n_regions=2)

    for invalid_threshold in ['80%', 'auto', -1.0]:
        with pytest.raises(
                ValueError,
                match="threshold given as ratio to the number of voxels must "
                      "be Real number and should be positive "
                      "and between 0 and total number of maps "
                      "i.e. n_maps={0}. "
                      "You provided {1}".format(maps.shape[-1],
                                                invalid_threshold)):
            _threshold_maps_ratio(maps, threshold=invalid_threshold)


def test_nans_threshold_maps_ratio():
    maps, _ = generate_maps((10, 10, 10), n_regions=2)
    data = get_data(maps)
    data[:, :, 0] = np.nan

    maps_img = nibabel.Nifti1Image(data, np.eye(4))
    thr_maps = _threshold_maps_ratio(maps_img, threshold=0.8)


def test_threshold_maps_ratio():
    # smoke test for function _threshold_maps_ratio with randomly
    # generated maps

    rng = np.random.RandomState(42)

    maps, _ = generate_maps((6, 8, 10), n_regions=3)

    # test that there is no side effect
    get_data(maps)[:3] = 100
    maps_data = get_data(maps).copy()
    thr_maps = _threshold_maps_ratio(maps, threshold=1.0)
    np.testing.assert_array_equal(get_data(maps), maps_data)

    # make sure that n_regions (4th dimension) are kept same even
    # in thresholded image
    assert thr_maps.shape[-1] == maps.shape[-1]

    # check that the size should be same for 3D image
    # before and after thresholding
    img = np.zeros((30, 30, 30)) + 0.1 * rng.standard_normal(size=(30, 30, 30))
    img = nibabel.Nifti1Image(img, affine=np.eye(4))
    thr_maps_3d = _threshold_maps_ratio(img, threshold=0.5)
    assert img.shape == thr_maps_3d.shape


def test_invalids_extract_types_in_connected_regions():
    maps, _ = generate_maps((10, 11, 12), n_regions=2)
    valid_names = ['connected_components', 'local_regions']

    # test whether same error raises as expected when invalid inputs
    # are given to extract_type in connected_regions function
    message = ("'extract_type' should be {0}")
    for invalid_extract_type in ['connect_region', 'local_regios']:
        with pytest.raises(ValueError, match=message.format(valid_names)):
            connected_regions(maps, extract_type=invalid_extract_type)


def test_connected_regions():
    rng = np.random.RandomState(42)

    # 4D maps
    n_regions = 4
    maps, mask_img = generate_maps((30, 30, 30), n_regions=n_regions)
    # 3D maps
    map_img = np.zeros((30, 30, 30)) + 0.1 * rng.standard_normal(
        size=(30, 30, 30)
    )
    map_img = nibabel.Nifti1Image(map_img, affine=np.eye(4))

    # smoke test for function connected_regions and also to check
    # if the regions extracted should be equal or more than already present.
    # 4D image case
    for extract_type in ['connected_components', 'local_regions']:
        connected_extraction_img, index = connected_regions(maps, min_region_size=10,
                                                            extract_type=extract_type)
        assert connected_extraction_img.shape[-1] >= n_regions
        assert index, np.ndarray
        # For 3D images regions extracted should be more than equal to one
        connected_extraction_3d_img, _ = connected_regions(map_img, min_region_size=10,
                                                           extract_type=extract_type)
        assert connected_extraction_3d_img.shape[-1] >= 1

    # Test input mask_img
    mask = get_data(mask_img)
    mask[1, 1, 1] = 0
    extraction_with_mask_img, index = connected_regions(maps,
                                                        mask_img=mask_img)
    assert extraction_with_mask_img.shape[-1] >= 1

    extraction_without_mask_img, index = connected_regions(maps)
    assert np.all(get_data(extraction_with_mask_img)[mask == 0] == 0.)
    assert not np.all(get_data(extraction_without_mask_img)[mask == 0] == 0.)

    # mask_img with different shape
    mask = np.zeros(shape=(10, 11, 12), dtype=np.int)
    mask[1:-1, 1:-1, 1:-1] = 1
    affine = np.array([[2., 0., 0., 0.],
                       [0., 2., 0., 0.],
                       [0., 0., 2., 0.],
                       [0., 0., 0., 2.]])
    mask_img = nibabel.Nifti1Image(mask, affine=affine)
    extraction_not_same_fov_mask, _ = connected_regions(maps,
                                                        mask_img=mask_img)
    assert maps.shape[:3] == extraction_not_same_fov_mask.shape[:3]
    assert mask_img.shape != extraction_not_same_fov_mask.shape[:3]

    extraction_not_same_fov, _ = connected_regions(maps)
    assert (np.sum(get_data(extraction_not_same_fov) == 0) >
                   np.sum(get_data(extraction_not_same_fov_mask) == 0))


def test_invalid_threshold_strategies():
    maps, _ = generate_maps((6, 8, 10), n_regions=1)

    extract_strategy_check = RegionExtractor(maps, thresholding_strategy='n_')
    valid_strategies = ['ratio_n_voxels', 'img_value', 'percentile']
    with pytest.raises(ValueError,
                       match="'thresholding_strategy' should be either of "
                             "these".format(valid_strategies)):
        extract_strategy_check.fit()


def test_threshold_as_none_and_string_cases():
    maps, _ = generate_maps((6, 8, 10), n_regions=1)

    extract_thr_none_check = RegionExtractor(maps, threshold=None)
    with pytest.raises(ValueError,
                       match="The given input to threshold is not valid."):
        extract_thr_none_check.fit()
    extract_thr_string_check = RegionExtractor(maps, threshold='30%')
    with pytest.raises(ValueError,
                       match="The given input to threshold is not valid."):
        extract_thr_string_check.fit()


def test_region_extractor_fit_and_transform():
    n_regions = 9
    n_subjects = 5
    maps, mask_img = generate_maps((40, 40, 40), n_regions=n_regions)

    # Test maps are zero in the mask
    mask_data = get_data(mask_img)
    mask_data[1, 1, 1] = 0
    extractor_without_mask = RegionExtractor(maps)
    extractor_without_mask.fit()
    extractor_with_mask = RegionExtractor(maps, mask_img=mask_img)
    extractor_with_mask.fit()
    assert not np.all(
        get_data(extractor_without_mask.regions_img_)[mask_data == 0] == 0.)
    assert np.all(
        get_data(extractor_with_mask.regions_img_)[mask_data == 0] == 0.)

    # smoke test to RegionExtractor with thresholding_strategy='ratio_n_voxels'
    extract_ratio = RegionExtractor(maps, threshold=0.2,
                                    thresholding_strategy='ratio_n_voxels')
    extract_ratio.fit()
    assert extract_ratio.regions_img_ != ''
    assert extract_ratio.regions_img_.shape[-1] >= 9

    # smoke test with threshold=string and strategy=percentile
    extractor = RegionExtractor(maps, threshold=30,
                                thresholding_strategy='percentile',
                                mask_img=mask_img)
    extractor.fit()
    assert extractor.index_, np.ndarray
    assert extractor.regions_img_ != ''
    assert extractor.regions_img_.shape[-1] >= 9

    n_regions_extracted = extractor.regions_img_.shape[-1]
    shape = (91, 109, 91, 7)
    expected_signal_shape = (7, n_regions_extracted)
    for id_ in range(n_subjects):
        img, data = _make_random_data(shape)
        # smoke test NiftiMapsMasker transform inherited in Region Extractor
        signal = extractor.transform(img)
        assert expected_signal_shape == signal.shape

    # smoke test with high resolution image
    maps, mask_img = generate_maps((20, 20, 20), n_regions=n_regions,
                                   affine=.2 * np.eye(4))

    extract_ratio = RegionExtractor(maps,
                                    thresholding_strategy='ratio_n_voxels',
                                    smoothing_fwhm=.6,
                                    min_region_size=.4)
    extract_ratio.fit()
    assert extract_ratio.regions_img_ != ''
    assert extract_ratio.regions_img_.shape[-1] >= 9

    # smoke test with zeros on the diagonal of the affine
    affine = np.eye(4)
    affine[[0, 1]] = affine[[1, 0]]  # permutes first and second lines
    maps, mask_img = generate_maps((40, 40, 40), n_regions=n_regions,
                                   affine=affine)

    extract_ratio = RegionExtractor(maps, threshold=0.2,
                                    thresholding_strategy='ratio_n_voxels')
    extract_ratio.fit()
    assert extract_ratio.regions_img_ != ''
    assert extract_ratio.regions_img_.shape[-1] >= 9


def test_error_messages_connected_label_regions():
    shape = (13, 11, 12)
    affine = np.eye(4)
    n_regions = 2
    labels_img = generate_labeled_regions(shape, affine=affine,
                                          n_regions=n_regions)
    with pytest.raises(
            ValueError,
            match="Expected 'min_size' to be specified as integer."):
        connected_label_regions(labels_img=labels_img, min_size='a')
    with pytest.raises(
            ValueError,
            match="'connect_diag' must be specified as True or False."):
        connected_label_regions(labels_img=labels_img, connect_diag=None)


def test_remove_small_regions():
    data = np.array([[[0., 1., 0.],
                      [0., 1., 1.],
                      [0., 0., 0.]],
                     [[0., 0., 0.],
                      [1., 0., 0.],
                      [0., 1., 0.]],
                     [[0., 0., 1.],
                      [1., 0., 0.],
                      [0., 1., 1.]]])
    # To remove small regions, data should be labelled
    label_map, n_labels = ndimage.label(data)
    sum_label_data = np.sum(label_map)

    affine = np.eye(4)
    min_size = 10
    # data can be act as mask_data to identify regions in label_map because
    # features in label_map are built upon non-zeros in data
    index = np.arange(n_labels + 1)
    removed_data = _remove_small_regions(label_map, index, affine, min_size)
    sum_removed_data = np.sum(removed_data)

    assert sum_removed_data < sum_label_data


def test_connected_label_regions():
    shape = (13, 11, 12)
    affine = np.eye(4)
    n_regions = 9
    labels_img = generate_labeled_regions(shape, affine=affine,
                                          n_regions=n_regions)
    labels_data = get_data(labels_img)
    n_labels_wo_reg_ext = len(np.unique(labels_data))

    # region extraction without specifying min_size
    extracted_regions_on_labels_img = connected_label_regions(labels_img)
    extracted_regions_labels_data = get_data(extracted_regions_on_labels_img)
    n_labels_wo_min = len(np.unique(extracted_regions_labels_data))

    assert n_labels_wo_reg_ext < n_labels_wo_min

    # with specifying min_size
    extracted_regions_with_min = connected_label_regions(labels_img,
                                                         min_size=100)
    extracted_regions_with_min_data = get_data(extracted_regions_with_min)
    n_labels_with_min = len(np.unique(extracted_regions_with_min_data))

    assert n_labels_wo_min > n_labels_with_min

    # Test connect_diag=False
    ext_reg_without_connect_diag = connected_label_regions(labels_img,
                                                           connect_diag=False)
    data_wo_connect_diag = get_data(ext_reg_without_connect_diag)
    n_labels_wo_connect_diag = len(np.unique(data_wo_connect_diag))
    assert n_labels_wo_connect_diag > n_labels_wo_reg_ext

    # If min_size is large and if all the regions are removed then empty image
    # will be returned
    extract_reg_min_size_large = connected_label_regions(labels_img,
                                                         min_size=500)
    assert np.unique(get_data(extract_reg_min_size_large)) == 0

    # Test the names of the brain regions given in labels.
    # Test labels for 9 regions in n_regions
    labels = ['region_a', 'region_b', 'region_c', 'region_d', 'region_e',
              'region_f', 'region_g', 'region_h', 'region_i']

    # If labels are provided, first return will contain extracted labels image
    # and second return will contain list of new names generated based on same
    # name with assigned on both hemispheres for example.
    extracted_reg, new_labels = connected_label_regions(labels_img,
                                                        min_size=100,
                                                        labels=labels)
    # The length of new_labels returned can differ depending upon min_size. If
    # min_size given is more small regions can be removed therefore newly
    # generated labels can be less than original size of labels. Or if min_size
    # is less then newly generated labels can be more.

    # We test here whether labels returned are empty or not.
    assert new_labels != ''
    assert len(new_labels) <= len(labels)

    # labels given in numpy array
    labels = np.asarray(labels)
    extracted_reg2, new_labels2 = connected_label_regions(labels_img,
                                                          labels=labels)
    assert new_labels != ''
    # By default min_size is less, so newly generated labels can be more.
    assert len(new_labels2) >= len(labels)

    # If number of labels provided are wrong (which means less than number of
    # unique labels in labels_img), then we raise an error

    # Test whether error raises
    unique_labels = set(np.unique(np.asarray(get_data(labels_img))))
    unique_labels.remove(0)

    # labels given are less than n_regions=9
    provided_labels = ['region_a', 'region_c', 'region_f',
                       'region_g', 'region_h', 'region_i']

    assert len(provided_labels) < len(unique_labels)

    with pytest.raises(ValueError):
        connected_label_regions(labels_img, labels=provided_labels)

    # Test if unknown/negative integers are provided as labels in labels_img,
    # we raise an error and test the same whether error is raised.
    # Introduce data type of float, see issue: https://github.com/nilearn/nilearn/issues/2580
    labels_data = np.zeros(shape, dtype=np.float32)
    h0 = shape[0] // 2
    h1 = shape[1] // 2
    h2 = shape[2] // 2
    labels_data[:h0, :h1, :h2] = 1
    labels_data[:h0, :h1, h2:] = 2
    labels_data[:h0, h1:, :h2] = 3
    labels_data[:h0, h1:, h2:] = -4
    labels_data[h0:, :h1, :h2] = 5
    labels_data[h0:, :h1, h2:] = 6
    labels_data[h0:, h1:, :h2] = np.nan
    labels_data[h0:, h1:, h2:] = np.inf

    neg_labels_img = nibabel.Nifti1Image(labels_data, affine)
    with pytest.raises(ValueError):
        connected_label_regions(labels_img=neg_labels_img)

    # If labels_img provided is 4D Nifti image, then test whether error is
    # raised or not. Since this function accepts only 3D image.
    labels_4d_data = np.zeros((shape) + (2, ))
    labels_data[h0:, h1:, :h2] = 0
    labels_data[h0:, h1:, h2:] = 0
    labels_4d_data[..., 0] = labels_data
    labels_4d_data[..., 1] = labels_data
    labels_img_4d = nibabel.Nifti1Image(labels_4d_data, np.eye(4))
    with pytest.raises(DimensionError):
        connected_label_regions(labels_img=labels_img_4d)

    # Test if labels (or names to regions) given is a string without a list.
    # Then, we expect it to be split to regions extracted and returned as list.
    labels_in_str = 'region_a'
    labels_img_in_str = generate_labeled_regions(shape, affine=affine,
                                                 n_regions=1)
    extract_regions, new_labels = connected_label_regions(labels_img_in_str,
                                                          labels=labels_in_str)
    assert isinstance(new_labels, list)

    # If user has provided combination of labels, then function passes without
    # breaking and new labels are returned based upon given labels and should
    # be equal or more based on regions extracted
    combined_labels = ['region_a', '1', 'region_b', '2', 'region_c', '3',
                       'region_d', '4', 'region_e']
    ext_reg, new_labels = connected_label_regions(labels_img,
                                                  labels=combined_labels)
    assert len(new_labels) >= len(combined_labels)
